import json
import pytest
import glog as log
from multivitamin.data.request import Request


r1 = """{"url":"https://upload.wikimedia.org/wikipedia/commons/thumb/1/1e/A_blank_black_picture.jpg/1536px-A_blank_black_picture.jpg",
    "dst_url":"https://dummy.url",
    "prev_response":"","bin_encoding":"true","bin_decoding":"true"}"""
r2 = """{"url":"http://pbs.twimg.com/media/*&&jesMessageWeird&&f---Char:large"}"""
r3 = "faulty_message"


def test_r1():
    req = Request(json.loads(r1))
    assert (
        req.url
        == "https://upload.wikimedia.org/wikipedia/commons/thumb/1/1e/A_blank_black_picture.jpg/1536px-A_blank_black_picture.jpg"
    )


def test_r2():
    # test standardize_url
    pass


def test_r3():
    with pytest.raises(ValueError):
        Request(json.loads(r3))



import os
import sys
import glog as log
import boto3
import json
<<<<<<< HEAD
from multivitamin.avro_api.frame_drawer import FrameDrawer
from multivitamin.avro_api.avro_api import AvroIO, AvroAPI

S3_BUCKET = "cv-applications-responses-staging"
# JSON_NAME="Replay%20Video%20Capture_2018-11-16_11.52.51-2816an1tb0v"
# JSON_NAME="538_Pelicans+vs+Thunder+11%3A5-fhj713lbrhi.30-31"
JSON_NAME = "1%3A3%20Houston%20Rockets%20at%20Golden%20State%20Warriors-6tgm4my1dr6"
S3_KEY = "SPORTS-NBALeagueDetector_1.0.0/" + JSON_NAME + ".json"


def test_FrameDrawer():
    S3 = boto3.resource("s3")
    file = S3.Object(S3_BUCKET, S3_KEY)
    AVRO_JSON = json.loads(file.get()["Body"].read().decode("utf-8"))
    output_folder = "./frame_visualization/" + JSON_NAME
    try:
        os.makedirs(output_folder)
    except:
        pass
    avro_api = AvroAPI(doc=AVRO_JSON)
    fd = FrameDrawer(avro_api=avro_api, decode=False, dump=True, out=output_folder)
    fd.process()
=======
from vitamincv.applications.general.frame_drawer import FrameDrawer
from vitamincv.avro_api.avro_api import AvroIO, AvroAPI

S3_BUCKET = "vitamincv-data"
JSON_NAME_1="Winnipeg%2520Jets%2520%2540%2520St.%2520Louis%2520Blues-uvr287fay9k-1min"
S3_KEY_1 = "jsons/" + JSON_NAME_1 + ".json"


def test_FrameDrawer():
    S3 = boto3.resource('s3')
    file = S3.Object(S3_BUCKET, S3_KEY_1)
    AVRO_JSON = json.loads(file.get()['Body'].read().decode('utf-8'))
    output_folder='./frame_drawer_output/'  
    avro_api=AvroAPI(doc=AVRO_JSON )
    fd = FrameDrawer(avro_api=avro_api,pushing_folder=output_folder,s3_bucket='vitamincv-data',s3_key='frame_drawer')    
    fd.process(dump_video=True)
def test_FrameDrawer_B():
    S3 = boto3.resource('s3')
    file = S3.Object(S3_BUCKET, S3_KEY_1)
    AVRO_JSON = json.loads(file.get()['Body'].read().decode('utf-8'))
    output_folder='./frame_drawer_output/'
    avro_api=AvroAPI(doc=AVRO_JSON )
    fd = FrameDrawer(avro_api=avro_api,s3_bucket='vitamincv-data',s3_key='frame_drawer')    
    fd.process(dump_images=True)
>>>>>>> origin/develop



import os
import json
import traceback
import threading

import glog as log
from flask import Flask, jsonify

from .apis import CommAPI
from .module import Module
from .data import Response, Request


HEALTHPORT = os.environ.get("PORT", 5000)


class Server(Flask):
    def __init__(
        self,
        modules,
        input_comm,
        output_comms=None,
        schema_registry_url=None,
    ):
        """Serves as the public interface for CV services through multivitamin.

        It's role is to start the healthcheck endpoint and initiate the
        services.

        Args:
            modules (list[Module]): list of concrete child implementation
                                    of Module
            input_comm (CommAPI): Concrete child implementation of CommAPI,
                                  called for pulling responses to process
            output_comms (list[CommAPI]): List of concrete child
                                          implementations of CommAPI,
                                          called for pushing responses to
                                          somewhere
            schema_registry_url (str): use schema in registry url instead of
                                       local schema
        """
        if isinstance(modules, Module):
            modules = [modules]
        assert(isinstance(modules, list))
        if len(modules) == 0:
            raise ValueError("No modules was provided.")
        for m in modules:
            assert(isinstance(m, Module))
        if not input_comm:
            raise TypeError("No input_comm set")
        if not output_comms:
            output_comms = [input_comm]
        if not isinstance(output_comms, list):
            output_comms = [output_comms]
        for out in output_comms:
            assert(isinstance(out, CommAPI))

        self.input_comm = input_comm
        self.output_comms = output_comms
        self.modules_info = [
            {"name": mod.name, "version": mod.version} for mod in modules
        ]
        self.modules = modules
        self.schema_registry_url = schema_registry_url

        log.info("Input comm type: {}".format(type(input_comm)))
        for out in output_comms:
            log.info("Output comm type(s): {}".format(type(out)))

        super().__init__(__name__)

        @self.route("/health", methods=["GET"])
        def health_check():
            return jsonify(self.modules_info)

    def start(self):
        """Public entry point for starting a server.

        Starts a healthcheck endpoint in a separate thread and calls
        self._start() which runs the actual server, pulling, processing,
        and posting requests
        """
        log.info(
            f"Starting HealthCheck endpoint at /health on port {HEALTHPORT}"
        )
        try:
            threading.Thread(
                target=self.run,
                kwargs={"host": "0.0.0.0", "port": HEALTHPORT},
                daemon=True,
            ).start()
        except Exception as e:
            log.error(e)
            log.error(traceback.print_exc())
            log.error("Error setting up healthcheck endpoint")
        log.info("Starting server...")
        self._start()

    def _start(self):
        """Start server. While loop that pulls requests from the input_comm,
        calls _process_request(request), and posts responses to output_comms
        """
        while True:
            try:
                log.debug("Pulling requests")
                requests = self.input_comm.pull()
                for request in requests:
                    try:
                        if request.kill_flag is True:
                            log.info(
                                "Incoming request with kill_flag == True,"
                                " killing server"
                            )
                            return
                        response = self._process_request(request)
                        log.info(
                            f"Pushing reponse for {request} to output_comms"
                        )
                        for output_comm in self.output_comms:
                            try:
                                ret = output_comm.push(response)
                            except Exception as e:
                                log.error(e)
                                log.error(traceback.format_exc())
                                log.error(
                                    f"Error pushing response for {request}"
                                    f" to output_comm: {output_comm}"
                                )
                    except Exception:
                        log.error(traceback.format_exc())
                        log.error(f"Error processing request: {request}")
            except Exception as e:
                log.error(e)
                log.error(traceback.format_exc())
                log.error("Error processing requests")

    def _process_request(self, request):
        """Send request_message through all the modules

        Args:
            request (Request): incoming request

        Returns:
            Response: outgoing response message
        """
        if not isinstance(request, Request):
            raise ValueError(f"{request} is of type {type(request)},"
                             f" not {Request}")
        log.debug(f"Processing: {request}")

        response = Response(request, self.schema_registry_url)

        for module in self.modules:
            log.info(f"Processing {request} in module: {module}")
            response = module.process(response)
            log.info(f"Processing {request} in module: {module}"
                     f" ...Status: {module.code}")
            if response is not None:
                log.debug("response.to_dict():"
                          f" {json.dumps(response.to_dict(), indent=2)}")
        return response



import threading
from queue import Queue
import queue
import os
import string
import time
import glog as log

from rq import Queue as rQueue
from rq import Worker as rWorker
from redis import Redis
import shlex
from subprocess import Popen

import multiprocessing
from multiprocessing import Queue as mQueue


class ThreadWorker(threading.Thread):
    def __init__(self, q, func, timeout=1, on_kill=None, on_start=None):
        threading.Thread.__init__(self)
        log.info("New Worker Thread: {}".format(self.name))
        self._queue = q
        self._function = func
        self._timeout = timeout
        self._kill = False
        self._on_kill = on_kill
        self._on_start = on_start
        if self._on_start:
            self._on_start()

    def run(self):
        while not self._kill:
            try:
                log.debug("{} - Pulling from Queue".format(self.name))
                work = self._queue.get(timeout=self._timeout)
            except queue.Empty:
                log.debug("{} - Timed Out".format(self.name))
                continue
            if work:
                log.debug("{} - Running {}".format(self.name, self._function))
                log.debug("{} - Input: {}".format(self.name, work))
                self._function(work)

        if self._on_kill:
            self._on_kill()
        # print('Worker {} Dying'.format(os.getpid()))

    def kill(self):
        self._kill = True


class ThreadManager:
    def __init__(
        self, func, n=1, on_kill=None, on_start=None, timeout=1, max_queue_size=-1
    ):
        self.queue = Queue(max_queue_size)
        self._timeout = timeout
        self.workers = []
        for _ in range(n):
            self.workers.append(
                ThreadWorker(
                    self.queue,
                    func,
                    on_kill=on_kill,
                    on_start=on_start,
                    timeout=timeout,
                )
            )
            self.workers[-1].start()

    def _build_workers(self, func, n=1, on_kill=None, timeout=1, max_queue_size=-1):
        self.kill_workers(self.workers)
        self.workers = []
        self.queue = Queue(max_queue_size)
        for _ in range(n):
            self.workers.append(
                ThreadWorker(
                    self.queue,
                    func,
                    on_kill=on_kill,
                    on_start=on_start,
                    timeout=timeout,
                )
            )
            self.workers[-1].start()

    def kill_workers(self):
        for worker in self.workers:
            worker.kill()
            worker.join()

    def kill_workers_on_completion(self):
        while True:
            time.sleep(0.2)
            if self.queue.empty():
                break
        self.kill_workers()


class RedisWorker:
    def __init__(self, q, on_kill=None):
        self.queue = q
        self.on_kill = on_kill

    def run(self):
        start_worker_cmd = shlex.split("rq worker " + self.queue.name)
        self.process = Popen(start_worker_cmd)

    def kill(self):
        self.process.kill()
        if self.on_kill:
            self.on_kill()


class RedisManager:
    def __init__(self, n=1, on_kill=None, q_name="default-redis-queue"):
        redis_conn = Redis()
        self.queue = rQueue(q_name, connection=redis_conn)
        start_redis_server_cmd = shlex.split("redis-server")
        self.server = Popen(start_redis_server_cmd)
        self.workers = []
        for _ in range(n):
            self.workers.append(RedisWorker(self.queue, on_kill=on_kill))
            self.workers[-1].run()

    def kill_workers_on_completion(self):
        while True:
            time.sleep(0.2)
            if len(self.queue.job_ids) == 0:
                break
        self.kill_workers()

    def kill_workers(self):
        self.server.kill()
        for worker in self.workers:
            worker.kill()


class ProcessWorker(multiprocessing.Process):
    def __init__(self, q, func, timeout=1, on_kill=None, on_start=None):
        multiprocessing.Process.__init__(self)
        log.info("New Worker Process: {}".format(self.name))
        self._queue = q
        self._function = func
        self._timeout = timeout
        self._kill = False
        self._on_kill = on_kill
        self._on_start = on_start
        if self._on_start:
            self._on_start()

    def run(self):
        while not self._kill:
            try:
                log.debug("{} - Pulling from Queue".format(self.name))
                work = self._queue.get(timeout=self._timeout)
            except queue.Empty:
                log.debug("{} - Timed Out".format(self.name))
                continue
            if work:
                log.debug("{} - Running {}".format(self.name, self._function))
                log.debug("{} - Input: {}".format(self.name, work))
                self._function(work)

        if self._on_kill:
            self._on_kill()
        # print('Worker {} Dying'.format(os.getpid()))

    def kill(self):
        self._kill = True


class ProcessManager:
    def __init__(
        self, func, n=1, on_kill=None, on_start=None, timeout=1, max_queue_size=-1
    ):
        self.queue = mQueue(max_queue_size)
        self._timeout = timeout
        self.workers = []
        for _ in range(n):
            self.workers.append(
                ProcessWorker(
                    self.queue,
                    func,
                    on_kill=on_kill,
                    on_start=on_start,
                    timeout=timeout,
                )
            )
            self.workers[-1].start()

    def _build_workers(self, func, n=1, on_kill=None, timeout=1, max_queue_size=-1):
        self.kill_workers(self.workers)
        self.workers = []
        self.queue = mQueue(max_queue_size)
        for _ in range(n):
            self.workers.append(
                ProcessWorker(
                    self.queue,
                    func,
                    on_kill=on_kill,
                    on_start=on_start,
                    timeout=timeout,
                )
            )
            self.workers[-1].start()

    def kill_workers(self):
        for worker in self.workers:
            worker.kill()
            worker.join()

    def kill_workers_on_completion(self):
        while True:
            time.sleep(0.2)
            if self.queue.empty():
                break
        self.kill_workers()


def WorkerManager(
    func=None,
    n=1,
    on_kill=None,
    on_start=None,
    timeout=1,
    max_queue_size=-1,
    parallelization="thread",
    q_name="default-redis-queue",
):
    if parallelization == "thread":
        assert callable(func)
        return ThreadManager(
            func,
            n=n,
            on_kill=on_kill,
            on_start=on_start,
            timeout=timeout,
            max_queue_size=max_queue_size,
        )
    if parallelization == "redis":
        return RedisManager(n=n, on_kill=on_kill, q_name=q_name)
    if parallelization == "process":
        assert callable(func)
        return ProcessManager(
            func,
            n=n,
            on_kill=on_kill,
            on_start=on_start,
            timeout=timeout,
            max_queue_size=max_queue_size,
        )


if __name__ == "__main__":
    ###########
    # EXAMPLE #
    ###########
    q = Queue()
    workers = []
    for _ in range(2):
        workers.append(ThreadWorker(q, print))
        workers[-1].start()
    for char in string.printable:
        q.put(char)
    while True:
        time.sleep(0.2)
        if q.empty():
            break
    for worker in workers:
        worker.kill()
        worker.join()



import boto3

from io import BytesIO


def load_fileobj(s3_bucket_name, key):
    s3_client = boto3.client("s3")
    bytes_obj = BytesIO()
    s3_client.download_fileobj(s3_bucket_name, key, bytes_obj)
    return bytes_obj


def generate_fileobj_from_s3_folder(s3_bucket_name, s3_folder_key):
    s3 = boto3.resource("s3")
    s3_bucket = s3.Bucket(name=s3_bucket_name)
    for obj in s3_bucket.objects.filter(Prefix=s3_folder_key):
        if obj.key == s3_folder_key:
            continue
        bytes_obj = load_fileobj(obj.bucket_name, obj.key)
        yield obj.key, bytes_obj



import sys
import os
import json
import glog as log
from multivitamin.applications.classifiers.caffe_classifier import CaffeClassifier
import boto3
import zipfile

# request1='["com.amazon.sqs.javamessaging.MessageS3Pointer",{"s3BucketName":"cvapis-data","s3Key":"f2d9a5d0-06e6-4975-8889-3fe79508972b"}]'
request1 = '["com.amazon.sqs.javamessaging.MessageS3Pointer",{"s3BucketName":"cvapis-data","s3Key":"1be661e2-4190-4d05-a4a4-156ddf9967fc"}]'


message = ""


def test_classifier():
    log.info("Testing caffe classifier.")

    log.info("Downloading net_data")

    s3_client = boto3.client("s3")
    log.info("Downloading model.")
    tmp_filepath = "/tmp/net_data.zip"
    s3_client.download_file(
        "cvapis-data", "classifiers/nhllogoclassifier/net_data.zip", tmp_filepath
    )
    log.info("Model downloaded.")
    with open(tmp_filepath, "rb") as f:
        log.info("Unzipping it.")
        z = zipfile.ZipFile(f)
        for name in z.namelist():
            print("    Extracting file", name)
            z.extract(name, "/tmp/")
    log.info("Unzipped.")
    caffe_classifier = CaffeClassifier(
        server_name="CaffeClassifier", version="1.0", net_data_dir="/tmp/net_data/"
    )
    p = {}
    p["company"] = "gumgum"
    # p['property_type']='placement'
    caffe_classifier.set_prev_pois(prev_pois=[p])

    log.info("1---------------------")
    caffe_classifier.process(request1)
    caffe_classifier.update_response()
    log.info("Getting response")
    try:
        response = caffe_classifier.request_api.get_response()
    except Exception as e:
        # log.error(e)
        log.error("Problem retrieving response")
    log.info("Jsonifying response")
    response_json = json.dumps(response, indent=4)
    log.info("---------------------")
    log.info("---------------------")



import json
import traceback

import glog as log


DEFAULT_SAMPLE_RATE = 1.0


class Request:
    def __init__(self, request_input, request_id=None):
        """Data object to encapsulate and cleanse request

        Args:
            request_input (dict or str): input request json as str or dict
            request_id (str): ID tied to request (esp from AWS SQS)
        """
        d = {}
        if isinstance(request_input, str):
            try:
                d = json.loads(request_input)
            except ValueError as v:
                log.error(v)
                log.error(traceback.format_exc())
                raise ValueError("Error decoding str into dict")
        elif isinstance(request_input, dict):
            d = request_input
        else:
            raise ValueError(
                f"request input is type: {type(request_input)}, should be of type str or dict"
            )
        self.request = d
        self.request_id = request_id

    def get(self, key, default=None):
        """Get method for any arbitrary key. Allows user to set default val

        Args:
            key (str): key in request dict
            default (Any): default value if key not in dict

        Returns:
            Any: value
        """
        return self.request.get(key, default)

    @property
    def url(self):
        """Convenience getter for the request URLj

        Returns:
            str: url
        """
        return self.request.get("url")

    @property
    def sample_rate(self):
        """Convenience getter for the request's defined sample_rate

        Returns:
            float: sample_rate
        """
        return self.request.get("sample_rate", DEFAULT_SAMPLE_RATE)

    @property
    def bin_encoding(self):
        """Getter for bin_encoding flag. Defaults to False

        Returns:
            bool: bin_encoding flag
        """
        be = self.request.get("bin_encoding", False)
        if isinstance(be, str):
            be = be.lower() == "true"
        elif isinstance(be, int):
            be = be == 1
        return be

    @property
    def bin_decoding(self):
        """Getter for bin_decoding flag. Defaults to False

        Returns:
            bool: bin_decoding flag
        """
        de = self.request.get("bin_decoding", False)
        if isinstance(de, str):
            de = de.lower() == "true"
        elif isinstance(de, int):
            de = de == 1
        return de

    @property
    def base64_encoding(self):
        """Getter for base64_encoding flag. Defaults to False

        Returns:
            bool: base64_encoding flag
        """
        be = self.request.get("base64_encoding", False)
        if isinstance(be, str):
            be = be.lower() == "true"
        elif isinstance(be, int):
            be = be == 1
        return be

    @property
    def prev_response(self):
        """Getter for prev_response

        Returns:
            str: prev_response
        """
        return self.request.get("prev_response")

    @prev_response.setter
    def prev_response(self, prev_response):
        """Getter for prev_response

        Returns:
            str: prev_response
        """
        self.request["prev_response"] = prev_response

    @property
    def prev_response_url(self):
        """Getter for prev_response_url

        Returns:
            str: prev_response_url
        """
        return self.request.get("prev_response_url")

    @property
    def dst_url(self):
        """Getter for destination url for response

        Returns:
            str: destination url
        """
        return self.request.get("dst_url")

    @property
    def flags(self):
        """Getter for arbitrary flags

        Returns:
            dict: flags
        """
        return self.request.get("flags")

    @property
    def kill_flag(self):
        """Getter for kill flag

        Returns:
            bool: kill flag
        """
        kf = self.request.get("kill_flag", False)
        if isinstance(kf, str):
            kf = kf.lower() == "true"
        elif isinstance(kf, int):
            kf = kf == 1
        return kf

    def __repr__(self):
        return f"Request({self.request}) <ID {self.request_id}>"



import os
import json
import itertools as it

import glog as log


def load_idmap(idmap_filepath):
    """Load idmap
    
    Each row corresponds to an int_id, label_str pairing, tab-separated    
    
    Args:
        idmap_filepath (str): path to idmap
    
    Raises:
        FileNotFoundError: if idmap does not exist
        ValueError: Problems parsing 
    """
    idmap = {}
    if not os.path.exists(idmap_filepath):
        raise FileNotFoundError("{} not found.".format(idmap_filepath))
    try:
        with open(idmap_filepath, "r") as rf:
            for row_in in rf:
                try:
                    row = row_in.split("\t")
                    assert len(row) >= 2
                except:
                    row = row_in.split(" ")
                    assert len(row) >= 2
                id = int(row[0])
                label = row[1].strip()
                idmap[label] = id
    except:
        log.error("Problems parsing " + idmap_filepath)
        raise ValueError("Problems parsing " + idmap_filepath)
    return idmap


def min_conf_filter_predictions(filter_dict, preds, confs, label_dict=None):
    """ Filter our predictions based on per label confidence thresholds

    Args:
        filter_dict (dict): A dict of strings or ints that get mapped a minimum confidence 
                            value. If the keys are strings, label_dict is required.
        preds (list): A list of predicted labels (strings) or classes (ints)
        confs (list): A list of floats associated to confidence 
                        values for each predicted class 
        label_dict (dict): A dict mapping the classifier output index to a string

    Returns:
        qualifying_preds (list): A list of elements from preds 
                                    that have qualifying confidence
    """
    if label_dict is None:
        label_dict = {}

    qualifying_preds = []
    for pred, conf in zip(preds, confs):
        min_conf = filter_dict.get(pred)
        if min_conf is None:
            min_conf = filter_dict.get(label_dict.get(pred))

        if min_conf is None:
            min_conf = 0
        if conf >= min_conf:
            qualifying_preds.append(pred)
    return qualifying_preds


def pandas_query_matches_props(bool_exp, props):
    """Evaluates the boolean expression on a list of properties
    
        Note: list of properties must be a pandas.DataFrame
        See: https://pandas.pydata.org/pandas-docs/stable/reference/api/pandas.DataFrame.query.html
    Args:
        bool_exp (str): boolean expression
        props (pd.DataFrame): DataFrame containing list of dicts (properties)
    
    Returns:
        bool: True if bool exp evals to True, else False
    """
    log.debug(f"bool_exp: {bool_exp}")
    log.debug(f"against properties: {json.dumps(props.to_dict(), indent=2)}")
    queried_pois = props.query(bool_exp)
    log.debug(f"matches? : {not queried_pois.empty}")
    return not queried_pois.empty


def convert_props_to_pandas_query(query_props):
    """Convert a list of query dicts to a boolean expression to be used in pandas.DataFrame.query

        E.g.
            [
                {"property_type":"object", "value":"face"}, 
                {"value":"car"}
            ]
            
            is converted to
            
            '(property_type == "object") & (value == "face") | (value == "car")'
    
    Args:
        query (list[dict]): list of dictionaries containing key/values of interest
    
    Returns:
        str: boolean expression
    """
    assert isinstance(query_props, list)
    log.debug(query_props)
    bool_exp = ""
    for i, q in enumerate(query_props):
        assert isinstance(q, dict)
        for j, (k, v) in enumerate(q.items()):
            bool_exp += f'({k} == "{v}")'
            if j != len(q) - 1:
                bool_exp += " & "
        if i != len(query_props) - 1:
            bool_exp += " | "
    log.debug(f"query: {query_props} transformed into boolean expression: {bool_exp}")
    return bool_exp


def batch_generator(iterator, batch_size):
# def batch_generator(iterable, batch_size):
    """Take an iterator, convert it to a batching generator

    See: https://realpython.com/python-itertools/

    Args:
        iterator: Any iterable object where each element is a list or a tuple of length N

    Returns:
        list: A list of N batches of size `self.batch_size`. The last
                batch may be smaller than the others
    """
    # iters = [iter(iterable)] * batch_size
    # return it.zip_longest(*iters, fillvalue=None)
    batch = []
    for iteration in iterator:
        batch.append(iteration)
        if len(batch) >= batch_size:
            yield zip(*batch)
            batch = []
    if len(batch) > 0:
        yield zip(*batch)



import os
import json
import glog as log
import tempfile
import shutil
import boto3
import hashlib

from multivitamin.apis.comm_api import CommAPI
from multivitamin.data import Response


INDENTATION = 2


class S3API(CommAPI):
    def __init__(self, s3_bucket, s3_key, bin_encoding=False):
        """ S3API is a CommAPI object that pushes responses to an
            s3_bucket/s3_keyget_fn or pushing_folder

        Args:
            s3_bucket (str): s3 bucket
            s3_key (str): key of s3 bucket to write to 
        """
        self.s3_bucket = s3_bucket
        self.s3_key = s3_key
        self.bin_encoding = bin_encoding

    def pull(self, n=1):
        raise NotImplementedError("S3API.pull() not yet implemented")

    def push(self, responses):
        """Push a list of Response objects to be written to pushing_folder

        Args:
            responses (list[Response]): list of requests to be pushed/written to disk
        """
        if not isinstance(responses, list):
            responses = [responses]

        for res in responses:
            assert isinstance(res, Response)
            fn = self.get_fn(res, self.bin_encoding)
            log.debug(f"fn: {fn}")
            tmp_dir = tempfile.mkdtemp()
            outfn = os.path.join(tmp_dir, fn)

            log.info("Writing response to {}".format(outfn))
            if not os.path.exists(os.path.dirname(outfn)):
                os.makedirs(os.path.dirname(outfn))

            if self.bin_encoding is True:
                with open(outfn, "wb") as wf:
                    wf.write(res.to_bytes())
            else:
                with open(outfn, "w") as wf:
                    wf.write(json.dumps(res.to_dict(), indent=INDENTATION))

            s3client = boto3.client("s3")
            assert(self.s3_key is not None)
            key_fullpath = os.path.join(self.s3_key, fn)
            log.info("Pushing {} to {}/{}".format(outfn, self.s3_bucket, key_fullpath))
            with open(outfn, "rb") as data:
                s3client.put_object(Bucket=self.s3_bucket, Key=key_fullpath, Body=data)

            if os.path.exists(tmp_dir):
                log.info("Removing temp dir {}".format(tmp_dir))
                shutil.rmtree(tmp_dir)

    def get_fn(self, response, bin_encoding):
        """Create a fn from url string from Response

        Args:
            response (Response): response
        
        Returns:
            str: unique identifier
        """
        media_url = response.url
        ext = ".json"
        if bin_encoding is True:
            ext = ".avro"
        hash = hashlib.md5(media_url.encode()).hexdigest()
        return os.path.basename(media_url) + "_" + hash +  ext



"""Adaptation of Martin Valgur's SeekableHTTPFile.

https://github.com/valgur/pyhttpio
"""

import cgi
import time
import urllib.request
from io import IOBase
from sys import stderr


class HTTPFile(IOBase):
    """Turns URLs into Filelike objects."""

    def __init__(self, url, name=None, repeat_time=-1, debug=False):
        """Allow a file accessible via HTTP to be used like a local file by
        utilities that use `seek()` to read arbitrary parts of the file, such
        as `ZipFile`. Seeking is done via the 'range: bytes=xx-yy' HTTP header.

        Parameters
        ----------
        url : str
            A HTTP or HTTPS URL
        name : str, optional
            The filename of the file.
            Will be filled from the Content-Disposition header if not provided.
        repeat_time : int, optional
            In case of HTTP errors wait `repeat_time` seconds before trying
            again.
            Negative value or `None` disables retrying and simply passes on
            the exception (the default).
        """
        super().__init__()
        self.url = url
        self.name = name
        self.repeat_time = repeat_time
        self.debug = debug
        self._pos = 0
        self._seekable = True
        with self._urlopen() as f:
            if self.debug:
                print(f.getheaders())
            self.content_length = int(f.getheader("Content-Length", -1))
            if self.content_length < 0:
                self._seekable = False
            if f.getheader("Accept-Ranges", "none").lower() != "bytes":
                self._seekable = False
            if name is None:
                header = f.getheader("Content-Disposition")
                if header:
                    value, params = cgi.parse_header(header)
                    self.name = params.get("filename")

    def seek(self, offset, whence=0):
        if not self.seekable():
            raise OSError
        if whence == 0:
            self._pos = 0
        elif whence == 1:
            pass
        elif whence == 2:
            self._pos = self.content_length
        self._pos += offset
        return self._pos

    def seekable(self, *args, **kwargs):
        return self._seekable

    def readable(self, *args, **kwargs):
        return not self.closed

    def writable(self, *args, **kwargs):
        return False

    def read(self, amt=-1):
        if self._pos >= self.content_length:
            return b""
        if amt < 0:
            end = self.content_length - 1
        else:
            end = min(self._pos + amt - 1, self.content_length - 1)
        byte_range = (self._pos, end)
        self._pos = end + 1
        with self._urlopen(byte_range) as f:
            return f.read()

    def readall(self):
        return self.read(-1)

    def tell(self):
        return self._pos

    def __getattribute__(self, item):
        attr = object.__getattribute__(self, item)
        if not object.__getattribute__(self, "debug"):
            return attr

        if hasattr(attr, '__call__'):
            def trace(*args, **kwargs):
                a = ", ".join(map(str, args))
                if kwargs:
                    a += ", ".join(["{}={}".format(k, v) for k, v in kwargs.items()])
                print("Calling: {}({})".format(item, a))
                return attr(*args, **kwargs)

            return trace
        else:
            return attr

    def _urlopen(self, byte_range=None):
        header = {}
        if byte_range:
            header = {"range": "bytes={}-{}".format(*byte_range)}
        while True:
            try:
                r = urllib.request.Request(self.url, headers=header)
                return urllib.request.urlopen(r)
            except urllib.error.HTTPError as e:
                if self.repeat_time is None or self.repeat_time < 0:
                    raise
                print("Server responded with " + str(e), file=stderr)
                print("Sleeping for {} seconds before trying again".format(self.repeat_time), file=stderr)
                time.sleep(self.repeat_time)



"""
example usage:
```
curl -d '{"url":"https://s3.amazonaws.com/vitamin-cv-test-data/data/media/images/nhl_image.jpg", "bin_encoding": "false"}' -H "Content-Type: application/json" -X POST http://localhost:8888/process
```
"""
import os

import boto3
import glog as log

from io import BytesIO
from multivitamin.http_server import HTTPServer
from multivitamin.applications.classifiers import CaffeClassifier


def load_fileobj(s3_bucket_name, key):
    s3_client = boto3.client("s3")
    bytes_obj = BytesIO()
    s3_client.download_fileobj(s3_bucket_name, key, bytes_obj)
    return bytes_obj


def generate_fileobj_from_s3_folder(s3_bucket_name, s3_folder_key):
    s3 = boto3.resource("s3")
    s3_bucket = s3.Bucket(name=s3_bucket_name)
    for obj in s3_bucket.objects.filter(Prefix=s3_folder_key):
        if obj.key == s3_folder_key:
            continue
        bytes_obj = load_fileobj(obj.bucket_name, obj.key)
        yield obj.key, bytes_obj


S3_BUCKET_NAME = "vitamin-cv-test-data"
S3_IMAGES_FOLDER = "data/media/images/generic/"
S3_EXPECTED_PREV_RESPONSES = "data/previous_responses"
S3_PREV_DETECTIONS_FOLDER = "data/sample_detections/random_boxes/"
S3_NET_DATA_FOLDER = "models/caffe/bvlc/SportsNHLLogoClassifier-v1.2/"  # DO NOT TOUCH

LOCAL_NET_DATA_DIR = "/tmp/test_caffe_classifier/net_data"


server_name = "TestClassifier"
version = "0.0.0"
cc = None
if os.path.exists(LOCAL_NET_DATA_DIR):
    try:
        cc = CaffeClassifier(server_name, version, LOCAL_NET_DATA_DIR)
    except Exception as e:
        log.error(e)
        pass
else:
    os.makedirs(LOCAL_NET_DATA_DIR)

for key, net_data_bytes in generate_fileobj_from_s3_folder(
    S3_BUCKET_NAME, S3_NET_DATA_FOLDER
):
    filename = os.path.basename(key)
    print("{}/{}".format(LOCAL_NET_DATA_DIR, filename))
    with open("{}/{}".format(LOCAL_NET_DATA_DIR, filename), "wb") as file:
        file.write(net_data_bytes.getvalue())

cc = CaffeClassifier(server_name, version, LOCAL_NET_DATA_DIR)

httpserver = HTTPServer([cc])
httpserver.start()



from multivitamin.module import Module


def test_prev_props_of_interest():
    class ConcreteModule(Module):
        def process(self, response):
            super().process(response)

    server_name = "Dummy"
    version = "1.0.0"
    cmod = ConcreteModule(server_name, version)
    pois = [{"property_type": "object", "value": "car"}]
    cmod.set_prev_props_of_interest(pois)



from abc import abstractmethod

import glog as log

from multivitamin.module import Module


class PropertiesModule(Module):
    def __init__(
        self,
        server_name,
        version,
        prop_type=None,
        prop_id_map=None,
        module_id_map=None,
    ):
        """Module that processes properties, not images
        
        Args:
            server_name (str): server_name
            version (str): version
            prop_type (str, optional): Defaults to None. property_type
            prop_id_map (str, optional): Defaults to None. Map: property_type -> id
            module_id_map (str, optional): Defaults to None. Map: server_name -> id
        """
        super().__init__(
            server_name=server_name,
            version=version,
            prop_type=prop_type,
            prop_id_map=prop_id_map,
            module_id_map=module_id_map,
        )
        log.debug("Creating PropertiesModule")

    def process(self, response):
        """Process method called by Server

        Args:
            response (Response): prev_response
        
        Returns:
            Response: resultant response
        """
        super().process(response)
        self.process_properties()
        return self.update_and_return_response()

    @abstractmethod
    def process_properties(self):
        """Abstract method to be implemented to the child PropertiesModule

        Should populate response with self.response.append_track(...) 
        or self.response.append_media_summary(...)
        """
        pass



import context
import sys
import os

import glog as log
import requests
import zipfile2
import boto3
from time import sleep

sys.path.append(os.path.abspath("../examples/ssd-caffe/"))
sys.path.append(os.path.abspath("../examples/"))
from dummy_detector import DummyDetector
from ssd_detector import SSDDetector
from multivitamin.module_api.server import Server
from multivitamin.comm_apis.sqs_api import SQSAPI
from multivitamin.comm_apis.vertex_api import VertexAPI
from multivitamin.comm_apis.local_api import LocalAPI

PORT = os.environ.get("PORT", 5000)

request1 = '{"url":"https://www.astrosafari.com/download/file.php?id=1608&sid=79f086b975b702945ca90b0ac3bd0202","dst_url":"https://vertex-api-v3.gumgum.com/v1/module-responses?featureId=5&moduleId=1&imageUrl\
=https%3A%2F%2Fwww.astrosafari.com%2Fdownload%2Ffile.php%3Fid%3D1608%26sid%3D79f086b975b702945ca90b0ac3bd0202&order=2&async=true","prev_response":"AAAAAEAGMi4wJjIwMTgtMTAtMTZUMTc6MjM6MjAAAAAAAAAAAAAAAhxOU0\
ZXQ2xhc3NpZmllcgAKNC4wLjEOU1VDQ0VTUwxndW1ndW0cMjAxODEwMTYxNzIzMjAObWFjaGluZQAAOk5TRldDbGFzc2lmaWVyXzIwMTgxMDE2MTcyMzIwALwBaHR0cHM6Ly93d3cuYXN0cm9zYWZhcmkuY29tL2Rvd25sb2FkL2ZpbGUucGhwP2lkPTE2MDgmYW1wO3NpZD0\
3OWYwODZiOTc1YjcwMjk0NWNhOTBiMGFjM2JkMDIwMrwBaHR0cHM6Ly93d3cuYXN0cm9zYWZhcmkuY29tL2Rvd25sb2FkL2ZpbGUucGhwP2lkPTE2MDgmYW1wO3NpZD03OWYwODZiOTc1YjcwMjk0NWNhOTBiMGFjM2JkMDIwMgCACsAHAgAAAAAChgEwLjAwMDBfKDAuMDAw\
MCwwLjAwMDApKDEuMDAwMCwwLjAwMDApKDEuMDAwMCwxLjAwMDApKDAuMDAwMCwxLjAwMDApAAgAAAAAAAAAAAAAgD8AAAAAAACAPwAAgD8AAAAAAACAPwACEgIcTlNGV0NsYXNzaWZpZXIACjQuMC4xDGd1bWd1bQxzYWZldHkIc2FmZQAc1nk/CtcjPAAAgD8AOk5TRldDb\
GFzc2lmaWVyXzIwMTgxMDE2MTcyMzIwAAAAAAIAAAAAAAAAAAISAhxOU0ZXQ2xhc3NpZmllchJoaXN0b2dyYW0KNC4wLjEMZ3VtZ3VtDHNhZmV0eQhzYWZlABzWeT8K1yM8AACAPwA6TlNGV0NsYXNzaWZpZXJfMjAxODEwMTYxNzIzMjAAAAAAAAA=","bin_encoding":"\
true","bin_decoding":"true"}'
request2 = '{"url":"https://s3.amazonaws.com/cvapis-media/images/Hackathon-Group.jpg","dst_url":"https://vertex-api-v3.gumgum.com/v1/module-responses?featureId=41&moduleId=1000&imageUrl=https://s3.amazonaws.com/cvapis-media/images/Hackathon-Group.jpg&order=1&async=true","prev_response":"","bin_encoding":"true","bin_decoding":"true"}'
request3 = '{"url":"https://s3.amazonaws.com/cvapis-media/images/Hackathon-Group.jpg","dst_url":"https://vertex-api-v3.gumgum.com/v1/module-responses?featureId=26&moduleId=1004&imageUrl=https://s3.amazonaws.com/cvapis-media/images/Hackathon-Group.jpg&order=1&async=true","prev_response":"","bin_encoding":"true","bin_decoding":"true"}'

message = ""

NET_DATA_DIR = "/tmp/net_data"
SERVER_NAME = "ObjectDetector"
VERSION = "1.0.0"
log.setLevel("INFO")


def pull_net_data():
    log.info("Downloading net_data")
    s3_client = boto3.client("s3")
    log.info("Downloading model.")
    tmp_filepath = "/tmp/net_data.zip"
    s3_client.download_file("cvapis-data", "ssd-detector/net_data.zip", tmp_filepath)
    log.info("Model downloaded.")
    with open(tmp_filepath, "rb") as f:
        log.info("Unzipping it.")
        z = zipfile2.ZipFile(f)
        for name in z.namelist():
            print("    Extracting file", name)
            z.extract(name, "/tmp/")
    log.info("Unzipped.")


def test_server_vertex_ssddetector():
    log.info("Testing a server with SSDDetector and SQSAPI")
    queue_name = "cvapis-testing"
    log.info("Sending test requests to queue named " + queue_name)
    sqs_api = SQSAPI(queue_name=queue_name)
    sqs_api.push([request1])
    sqs_api.push([request2])
    sqs_api.push([request3])
    log.info("Creating VertexAPI")
    vertex_api = VertexAPI(queue_name=queue_name)

    # if not os.path.exists(NET_DATA_DIR):
    if True:
        pull_net_data()

    ssd_detector = SSDDetector(SERVER_NAME, VERSION, net_data_dir=NET_DATA_DIR)

    server = Server(ssd_detector, vertex_api)
    log.info("Starting server")
    log.info("******************")
    log.info("******************")
    log.info("******************")
    server.start()


def test_server_localapi_ssddetector():
    log.info("Testing a server with SSDDetector and LocalAPI")

    pulling_folder = "data/requests"
    pushing_folder = "data/output"
    log.info("Folder of requests to process: {}".format(pulling_folder))

    # if not os.path.exists(NET_DATA_DIR):
    if True:
        pull_net_data()

    ssd_detector = SSDDetector(SERVER_NAME, VERSION, NET_DATA_DIR)

    local_api = LocalAPI(pulling_folder, pushing_folder, True)
    server = Server(ssd_detector, local_api)
    log.info("Starting server")
    server.start()


def test_vertexpoller_to_local_server():
    log.info("Testing a dummy server polling from a queue and dumping with LocalAPI")
    pushing_folder = "data/output_nhl_historical"
    dummy_detector = DummyDetector("dummy", "1.0", None)
    sqs_api = SQSAPI(queue_name="SPORTS-FinalResponses-DEV")
    local_api = LocalAPI(pulling_folder=None, pushing_folder=pushing_folder)
    server = Server(dummy_detector, sqs_api, [local_api])
    log.info("Starting server")
    server.start()


def test_health_endpoint():
    log.info("Testing a dummy server and querying the endpoint")
    dummy_detector = DummyDetector("dummy", "1.0", None)
    local_api = LocalAPI(pulling_folder=None, pushing_folder="output")
    server = Server(dummy_detector, local_api)
    log.info("Starting server")
    server.start()
    sleep(1)  # wait for server to start
    url = "http://0.0.0.0:{}/health".format(PORT)
    log.info("HTTP get on {}".format(url))
    res = requests.get(url)
    log.info("response.text: {}".format(res.text))
    assert res.status_code == 200
    assert res.text



import random
from collections import MutableMapping

from dataclasses import dataclass, field
from typing import List
from typeguard import typechecked

from multivitamin.data.response.utils import round_float

# Helper functions


def create_region_id():
    """Create a region_id

    Returns:
        str: 16 digit random number
    """
    return str(int(random.random() * 1e16))


def make_whole_image_contour():
    """Create a list of Points representing an entire 2D image
    
    Returns:
        List[Point]: upper left, upper right, bottom right, bottom left
    """
    return [Point(0.0, 0.0), Point(1.0, 0.0), Point(1.0, 1.0), Point(0.0, 1.0)]


def create_bbox_contour_from_points(
    xmin, ymin, xmax, ymax, bound=False, lb_x=0.0, ub_x=1.0, lb_y=0.0, ub_y=1.0
):
    """Helper function to create bounding box contour from 4 extrema points

    Args:
        xmin (float or int): point
        ymin (float or int): point
        xmax (float or int): point
        ymax (float or int): point
        bound (bool, optional): Defaults to False. Bound between lb and ub
        lb_x (float, optional): Defaults to 0.0.
        ub_x (float, optional): Defaults to 1.0.
        lb_y (float, optional): Defaults to 0.0.
        ub_y (float, optional): Defaults to 1.0.

    Returns:
        list[Point]: list of Points
    """
    if bound:
        xmin = max(lb_x, xmin)
        ymin = max(lb_y, ymin)
        xmax = min(ub_x, xmax)
        ymax = min(ub_y, ymax)
    return [Point(xmin, ymin), Point(xmax, ymin), Point(xmax, ymax), Point(xmin, ymax)]


# Data classes


class DictLike(MutableMapping):
    """Base class used for the below dataclasses, so that each dataclass can 
    act like a dict with [] access
    """

    def __init__(self, *args, **kwargs):
        self.__dict__.update(*args, **kwargs)

    def __setitem__(self, key, value):
        self.__dict__[key] = value

    def __getitem__(self, key):
        return self.__dict__[key]

    def __delitem__(self, key):
        del self.__dict__[key]

    def __iter__(self):
        return iter(self.__dict__)

    def __len__(self):
        return len(self.__dict__)

    def __str__(self):
        return str(self.__dict__)

    def __repr__(self):
        return f"{super().__repr__()}, ({self.__dict__})"


@typechecked
@dataclass
class PropPair(DictLike):
    property_type: str = ""
    value: str = ""


@typechecked
@dataclass
class Relationship(DictLike):
    relation: str = ""
    sources: List[str] = field(default_factory=list)


@typechecked
@dataclass
class EligibleProp(DictLike):
    server: str = ""
    property_type: str = ""
    value: str = ""
    confidence_min: float = 0.0
    father_properties: List[PropPair] = field(default_factory=list)


@typechecked
@dataclass
class AnnotationTask(DictLike):
    id: str = ""
    tstamps: List[float] = field(default_factory=list)
    labels: List[str] = field(default_factory=list)
    tags: List[EligibleProp] = field(default_factory=list)


@typechecked
@dataclass
class Footprint(DictLike):
    """Footprint of a module 

    Args:
        code (str): return code
        ver (str): version
        company (str): company
        labels (List[EligibleProp]): list of labels
        server_track (str): server_track
        date (str): date in YYYYMMDDhhmmss
        annotator (str): annotator
        tstamps (List[float]): list of timestamps processed by module
        id (str): footprint id
        num_images_processed (int): number of images processed
        request_source (str): source of requests, e.g. SQS queue name
    """

    code: str = ""
    ver: str = ""
    company: str = "gumgum"
    labels: List[EligibleProp] = field(default_factory=list)
    server_track: str = ""
    server: str = ""
    date: str = "20000101000000"
    annotator: str = ""
    tstamps: List[float] = field(default_factory=list)
    id: str = ""
    num_images_processed: int = 0
    request_source: str = ""


@typechecked
@dataclass
class Point(DictLike):
    """Point class

    Args:
        x (float): Defaults to 0.0
        y (float): Defaults to 0.0
    """

    x: float = 0.0
    y: float = 0.0


@typechecked
@dataclass
class Property(DictLike):
    """Property class

    Args:
        relationships (List[str]): N/A
        confidence (float): Defaults to 0.0
        confidence_min (float): Defaults to 0.0
        ver (str): version
        company (str): Defaults to "gumgum"
        server_track (str): server track
        value (str): value
        server (str): server/module name
        footprint_id (str): associated footprint
        fraction (float): Defaults to 0.0
        module_id (int): Defaults to 0, used for mapping server name to an ID
        property_type (str): type of the value field
        value_verbose (str): extra field for value
        property_id (int): mapping property value to an ID
    """

    relationships: List[str] = field(default_factory=list)
    confidence: float = 0.0
    confidence_min: float = 0.0
    ver: str = ""
    company: str = "gumgum"
    server_track: str = ""
    value: str = ""
    server: str = ""
    footprint_id: str = ""
    fraction: float = 0.0
    module_id: int = 0
    property_type: str = ""
    value_verbose: str = ""
    property_id: int = 0


@typechecked
@dataclass
class Region(DictLike):
    contour: List[Point] = field(default_factory=make_whole_image_contour)
    props: List[Property] = field(default_factory=list)
    father_id: str = ""
    features: str = ""
    id: str = field(default_factory=create_region_id)


@typechecked
@dataclass
class VideoAnn(DictLike):
    t1: float = 0.0
    t2: float = 0.0
    props: List[Property] = field(default_factory=list)
    regions: List[Region] = field(default_factory=list)
    region_ids: List[str] = field(default_factory=list)

    def __post_init__(self):
        self.t1 = round_float(self.t1)
        self.t2 = round_float(self.t2)


@typechecked
@dataclass
class ImageAnn(DictLike):
    t: float = 0.0
    regions: List[Region] = field(default_factory=list)

    def __post_init__(self):
        self.t = round_float(self.t)


@typechecked
@dataclass
class MediaAnn(DictLike):
    codes: List[Footprint] = field(default_factory=list)
    url_original: str = ""
    url: str = ""
    w: int = 0
    h: int = 0
    post_url: str = ""
    frames_annotation: List[ImageAnn] = field(default_factory=list)
    media_summary: List[VideoAnn] = field(default_factory=list)
    tracks_summary: List[VideoAnn] = field(default_factory=list)
    annotation_tasks: List[AnnotationTask] = field(default_factory=list)


@typechecked
@dataclass
class ResponseInternal(DictLike):
    point_aux: Point = None
    footprint_aux: Footprint = None
    proppair_aux: PropPair = None
    annotation_task_aux: AnnotationTask = None
    image_annotation_aux: ImageAnn = None
    video_annotation_aux: VideoAnn = None
    property_aux: Property = None
    relationship_aux: Relationship = None
    eligibleprop_aux: EligibleProp = None
    region_aux: Region = None
    version: str = ""
    date: str = "20000101000000"
    media_annotation: MediaAnn = field(default_factory=MediaAnn)



import warnings
from multivitamin.utils.GPUUtilities import GPUUtility, CupyUtility


def test_init():
    gpu_utility = GPUUtility()
    gpu_ids = gpu_utility.GPUs
    assert len(gpu_ids) >= 0


def test_get_gpus():
    gpu_utility = GPUUtility()
    gpu_ids = gpu_utility.get_gpus()
    assert len(gpu_ids) >= 0


def test_limit_property():
    gpu_utility = GPUUtility(limit=2.2)
    assert gpu_utility.limit == 2
    gpu_utility.limit = 3.4
    assert gpu_utility.limit == 3


def test_minFreeMemory_property():
    gpu_utility = GPUUtility(minFreeMemory=2.2)
    assert gpu_utility.minFreeMemory == 1
    gpu_utility.minFreeMemory = -2
    assert gpu_utility.minFreeMemory == 0
    gpu_utility.minFreeLoad = 0.5
    assert gpu_utility.minFreeLoad == 0.5


def test_minFreeLoad_property():
    gpu_utility = GPUUtility(minFreeLoad=2.2)
    assert gpu_utility.minFreeLoad == 1
    gpu_utility.minFreeLoad = -5
    assert gpu_utility.minFreeLoad == 0
    gpu_utility.minFreeLoad = 0.25
    assert gpu_utility.minFreeLoad == 0.25


def test_cupy_numpy_conversion():
    cupy_utility = CupyUtility(minFreeMemory=1.0)
    assert cupy_utility.xp.__name__ == "numpy"

    cupy_utility = CupyUtility()
    if not cupy_utility.xp.__name__ == "cupy":
        warnings.warn(
            RuntimeWarning(
                "Warning! No GPU with available load or memory. Verify first!"
            )
        )



import os
import shutil
import random
import traceback

import boto3
import glog as log
import cv2
from colour import Color

from multivitamin.module import PropertiesModule
from multivitamin.data import Response
from multivitamin.data.response.utils import p0p1_from_bbox_contour
from multivitamin.data.response.dtypes import Property, VideoAnn
from multivitamin.media import MediaRetriever
import hashlib


COLORS = [
    'darkorchid', 'darkgreen', 'coral', 'darkseagreen',
    'forestgreen', 'firebrick', 'olivedrab', 'steelblue',
    'tomato', 'yellowgreen'
]


def random_bgr():
    color = Color(COLORS[random.randint(0, len(COLORS)-1)]).rgb[::-1]
    return [x*255 for x in color]


def get_props_from_region(region):
    prop_strs = []
    for prop in region["props"]:
        try:
           prop["value"] = "{:.2f}".format(float(prop["value"]))
        except:
            pass
        out = "{}_{:.2f}".format(prop["value"], prop["confidence"])
        prop_strs.append(out)
    return prop_strs


DEFAULT_DUMP_FOLDER = '/tmp'


class FrameDrawer(PropertiesModule):
    def __init__(
        self,
        response=None,
        med_ret=None,
        server_name='FrameDrawer',
        version='1.0.0',
        module_id_map=None,
        pushing_folder=DEFAULT_DUMP_FOLDER,
        s3_bucket=None,
        s3_key=None
    ):
        super().__init__(
            server_name=server_name,
            version=version,
            module_id_map=module_id_map
        )

        if isinstance(response, dict):
            response = Response(response)

        self.response = response

        if not isinstance(self.response, Response):
            log.debug("No response set")
            self.response = None

        self.med_ret = med_ret
        if med_ret is None and self.response is not None:
            self.med_ret = MediaRetriever(self.response.url)

        if s3_bucket and not s3_key:
            raise ValueError("s3 bucket defined but s3_key not defined")

        if s3_key and not s3_key:
            raise ValueError("s3_key defined but s3 bucket not defined")

        if not pushing_folder and not s3_key:
            raise ValueError("pushing_folder and s3_key not defined, "
                             "we cannot set where to dump.")

        self.local_dir = pushing_folder
        self.s3_bucket = s3_bucket
        self.s3_key_prefix = s3_key

    def process_properties(self,
                           dump_video=True,
                           dump_images=False,
                           tstamps_of_interest=None):
        assert(isinstance(self.response, Response))
        self.med_ret = MediaRetriever(self.response.url)

        video_width, video_height = self.med_ret.get_w_h()
        media_id = os.path.basename(self.response.url).rsplit(".", 1)[0]

        media_id = "".join(
            [e for e in media_id if e.isalnum() or e in ["/", "."]]
        )
        content_type_map = {}

        dump_video = (self.request.get("dump_video") or dump_video) and \
            self.med_ret.is_video
        dump_images = (self.request.get("dump_images") or dump_images) or \
            self.med_ret.is_image

        if dump_images is False and self.med_ret.is_image:
            dump_images = True

        if dump_video is False and dump_images is False:
            log.warning("`dump_video` and `dump_images` are both false."
                        " Unable to proceed.")
            return

        log.debug(f"Dumping Video: {dump_video}")
        log.debug(f"Dumping Frames: {dump_images}")

        dump_folder = f"{self.local_dir}/"
        self.dumping_folder_url = dump_folder

        if dump_folder and not os.path.exists(dump_folder):
            os.makedirs(dump_folder)

        hash = hashlib.md5(self.med_ret.url.encode()).hexdigest()
        if dump_video:
            filename = os.path.join(dump_folder, f"{media_id}_{hash}.mp4")
            fps = 1
            frameSize = (video_width, video_height)
            fourcc = cv2.VideoWriter_fourcc(*'H264')

            log.debug("filename: " + filename)
            log.debug("fourcc: " + str(fourcc))
            log.debug("type(fourcc): " + str(type(fourcc)))
            log.debug("fps: " + str(fps))
            log.debug("type(fps): " + str(type(fps)))
            log.debug("frameSize: " + str(frameSize))
            log.debug("type(frameSize): " + str(type(frameSize)))

            vid = cv2.VideoWriter(filename, fourcc, fps, frameSize)
            content_type_map[os.path.basename(filename)] = 'video/mp4'
        elif dump_images and self.med_ret.is_image:
            filename = dump_folder + media_id + "_" + hash + ".jpeg"
        face = cv2.FONT_HERSHEY_SIMPLEX
        scale = 0.65
        thickness = 2

        # we get the image_annotation tstamps
        tstamps = self.response.get_timestamps()
        tstamp_frame_anns = self.response.get_timestamps_from_frames_ann()
        log.debug('tstamps: ' + str(tstamps))
        log.debug('tstamps_dets: ' + str(tstamp_frame_anns))

        # we get the frame iterator
        frames_iterator = []
        if tstamps_of_interest is not None:
            if type(tstamps_of_interest) is list:
                for t in tstamps_of_interest:
                    frame = self.med_ret.get_frame(tstamp=t)
                    frames_iterator.append((frame, t))
        else:
            try:
                frames_iterator = self.med_ret.get_frames_iterator(
                    sample_rate=1.0
                )
            except Exception:
                log.error(traceback.format_exc())
                raise Exception("Error loading media")

        for i, (img, tstamp) in enumerate(frames_iterator):
            if img is None or tstamp is None:
                m = f"frame at tstamp={tstamp}" if img is None else "tstamp"
                log.warning(f"Invalid {m}")
                continue

            if tstamp in tstamp_frame_anns:
                log.debug(f"drawing frame for tstamp: {tstamp}")
                regions = self.response.get_regions_from_tstamp(tstamp)
                for region in regions:
                    rand_color = random_bgr()
                    p0, p1 = p0p1_from_bbox_contour(
                        region['contour'], video_width, video_height
                    )
                    anchor_point = [p0[0] + 3, p1[1] - 3]

                    if abs(p1[1] - video_height) < 30:
                        anchor_point = [p0[0] + 3, int(p1[1] / 2) - 3]

                    img = cv2.rectangle(img, p0, p1, rand_color, thickness)

                    prop_strs = get_props_from_region(region)
                    for i, prop in enumerate(prop_strs):
                        img = cv2.putText(
                            img,
                            prop,
                            (anchor_point[0], anchor_point[1] + i * 25),
                            face,
                            1.0,
                            rand_color,
                            thickness
                        )
            elif tstamp in tstamps:
                log.debug("Making frame gray")
                gray = cv2.cvtColor(img, cv2.COLOR_BGR2GRAY)
                img = cv2.cvtColor(gray, cv2.COLOR_GRAY2BGR)
            else:
                log.debug("No processed frame")
                continue

            # Add timestamp to frame
            img = cv2.putText(
                img,
                str(tstamp),
                (20, 20),
                face, scale,
                [255, 255, 255],
                thickness
            )

            if dump_video:
                # we add the frame
                log.debug("Adding frame")
                vid.write(img)

            if dump_images:
                # we dump the frame
                if self.med_ret.is_video:
                    filename = os.path.join(
                        dump_folder,
                        f"{media_id}_{tstamp}.jpg"
                    )

                log.debug(f"Writing to file: {filename}")
                cv2.imwrite(filename, img)

                content_type_map[os.path.basename(filename)] = 'image/jpeg'

        if dump_video:
            vid.release()

        s3_url_root = None
        if self.s3_bucket:
            s3_url_root = self.upload_files(
                dump_folder, content_type_map, media_id
            )

        if self.local_dir == DEFAULT_DUMP_FOLDER:
            log.info('Removing files in ' + dump_folder)
            shutil.rmtree(dump_folder)

        props = []
        if dump_images:
            val = filename

            if self.med_ret.is_video:
                val = f"{dump_folder}/{media_id}_*.jpg"

            props.append(
                Property(
                    server=self.name,
                    ver=self.version,
                    value=val,
                    property_type="dumped_images",
                    property_id=1,
                )
            )
        if dump_video and s3_url_root is not None:
            dumped_video_url = os.path.basename(filename)
            props.append(
                Property(
                    server=self.name,
                    ver=self.version,
                    value=f"{s3_url_root}/{dumped_video_url}",
                    property_type="dumped_video",
                    property_id=2,
                )
            )
        media_summary = VideoAnn(t1=0.0, t2=tstamp, props=props)
        self.response.append_media_summary(media_summary)

    def upload_files(self, source, content_type_map, media_id):
        log.info("Uploading files")
        session = boto3.Session()
        s3 = session.resource('s3')
        bucket = s3.Bucket(self.s3_bucket)
        key_root = f"{self.s3_key_prefix}/{media_id}"
        # https://<bucket-name>.s3.amazonaws.com/<key>
        s3_url_root = \
            f"https://s3.amazonaws.com/{self.s3_bucket}/{key_root}"

        for subdir, dirs, files in os.walk(source):
            for file in files:
                full_path = os.path.join(subdir, file)
                with open(full_path, 'rb') as data:
                    rel_path = os.path.basename(full_path)
                    key = f"{key_root}/{rel_path}"
                    log.info('Pushing ' + full_path + ' to ' + s3_url_root)
                    try:
                        content_type = content_type_map[
                            os.path.basename(full_path)
                        ]
                    except Exception:
                        # File is not intended to be uploaded.
                        # It was not generated in this execution.
                        content_type = None

                    if content_type:
                        bucket.put_object(
                            Key=key, Body=data, ContentType=content_type
                        )

        return s3_url_root



import os


def load_idmap(idmap_file):
    """Load tab-separated idmap file containing label index and label string

    Args:
        idmap_file (str): filepath to idmap

    Returns:
        dict: labelmap (key=index, value=string)
    """
    if not os.path.exists(idmap_file):
        raise FileExistsError(idmap_file)

    labelmap = {}
    with open(idmap_file, "r") as rf:
        for row in rf:
            row = row.split("\t")
            labelmap[int(row[0])] = row[1].strip()
    return labelmap


def load_label_prototxt(prototxt_file):
    from google.protobuf import text_format
    from caffe.proto import caffe_pb2 as cpb2

    with open(prototxt_file) as f:
        labelmap_aux = cpb2.LabelMap()
        text_format.Merge(str(f.read()), labelmap_aux)
        num_labels = len(labelmap_aux.item)
        labelmap = {}
        for item in labelmap_aux.item:
            index = item.label
            label = item.display_name
            labelmap[index] = label
    return labelmap



import os
import json

import glog as log
import numpy as np
import boto3
import cv2

from multivitamin.applications.detectors.ssd_detector import SSDDetector
from multivitamin.data.request import Request
from multivitamin.data.response import Response
from multivitamin.data.response.io import load_response_from_request
from utils import generate_fileobj_from_s3_folder

S3_BUCKET_NAME = "vitamin-cv-test-data"
S3_IMAGES_FOLDER = "data/media/images/generic/"
S3_PREV_DETECTIONS_FOLDER = "data/sample_detections/random_boxes/"
S3_NET_DATA_FOLDER = "models/caffe/ssd/SportsNHLPlacementDetector-v1.2/"

import zipfile

request1 = '{"url":"https://www.astrosafari.com/download/file.php?id=1608&sid=79f086b975b702945ca90b0ac3bd0202","dst_url":"https://vertex-api-v3.gumgum.com/v1/module-responses?featureId=5&moduleId=1&imageUrl\
=https%3A%2F%2Fwww.astrosafari.com%2Fdownload%2Ffile.php%3Fid%3D1608%26sid%3D79f086b975b702945ca90b0ac3bd0202&order=2&async=true","prev_response":"AAAAAEAGMi4wJjIwMTgtMTAtMTZUMTc6MjM6MjAAAAAAAAAAAAAAAhxOU0\
ZXQ2xhc3NpZmllcgAKNC4wLjEOU1VDQ0VTUwxndW1ndW0cMjAxODEwMTYxNzIzMjAObWFjaGluZQAAOk5TRldDbGFzc2lmaWVyXzIwMTgxMDE2MTcyMzIwALwBaHR0cHM6Ly93d3cuYXN0cm9zYWZhcmkuY29tL2Rvd25sb2FkL2ZpbGUucGhwP2lkPTE2MDgmYW1wO3NpZD0\
3OWYwODZiOTc1YjcwMjk0NWNhOTBiMGFjM2JkMDIwMrwBaHR0cHM6Ly93d3cuYXN0cm9zYWZhcmkuY29tL2Rvd25sb2FkL2ZpbGUucGhwP2lkPTE2MDgmYW1wO3NpZD03OWYwODZiOTc1YjcwMjk0NWNhOTBiMGFjM2JkMDIwMgCACsAHAgAAAAAChgEwLjAwMDBfKDAuMDAw\
MCwwLjAwMDApKDEuMDAwMCwwLjAwMDApKDEuMDAwMCwxLjAwMDApKDAuMDAwMCwxLjAwMDApAAgAAAAAAAAAAAAAgD8AAAAAAACAPwAAgD8AAAAAAACAPwACEgIcTlNGV0NsYXNzaWZpZXIACjQuMC4xDGd1bWd1bQxzYWZldHkIc2FmZQAc1nk/CtcjPAAAgD8AOk5TRldDb\
GFzc2lmaWVyXzIwMTgxMDE2MTcyMzIwAAAAAAIAAAAAAAAAAAISAhxOU0ZXQ2xhc3NpZmllchJoaXN0b2dyYW0KNC4wLjEMZ3VtZ3VtDHNhZmV0eQhzYWZlABzWeT8K1yM8AACAPwA6TlNGV0NsYXNzaWZpZXJfMjAxODEwMTYxNzIzMjAAAAAAAAA=","bin_encoding":"\
true","bin_decoding":"true"}'
request2 = '{"url":"http://pbs.twimg.com/media/Dpq4c0OXcAII7J2.jpg:large","dst_url":"https://vertex-api-v3.gumgum.com/v1/module-responses?featureId=41&moduleId=1000&imageUrl=http%3A%2F%2Fpbs.twimg.com%2Fmedia%2FDpq4c0OXcAII7J2.jpg%3Alarge&order=1&async=true","prev_response":"","bin_encoding":"true","bin_decoding":"true"}'
request3 = '{"url":"http://pbs.twimg.com/media/Dpq40egWkAAtocH.jpg:large","dst_url":"https://vertex-api-v3.gumgum.com/v1/module-responses?featureId=26&moduleId=1004&imageUrl=http%3A%2F%2Fpbs.twimg.com%2Fmedia%2FDpq40egWkAAtocH.jpg%3Alarge&order=1&async=true","prev_response":"","bin_encoding":"true","bin_decoding":"true"}'

message = ""


def test_object_detector():
    log.info("Testing SSD detector.")
    log.info("Downloading net_data")
    s3_client = boto3.client("s3")
    log.info("Downloading model.")
    tmp_filepath = "/tmp/net_data.zip"
    s3_client.download_file(
        "cvapis-data",
        "ssd-detector/nhlplacementdetector/net_data_v1.1.zip",
        tmp_filepath,
    )
    log.info("Model downloaded.")
    with open(tmp_filepath, "rb") as f:
        log.info("Unzipping it.")
        z = zipfile.ZipFile(f)
        for name in z.namelist():
            print("    Extracting file", name)
            z.extract(name, "/tmp/")
    log.info("Unzipped.")
    ssd_detector = SSDDetector(
        server_name="SSDDetector", version="1.0", net_data_dir="/tmp/net_data/"
    )
    log.info("1---------------------")
    ssd_detector.process(request1)
    log.info("2---------------------")
    ssd_detector.process(request2)
    log.info("3---------------------")
    ssd_detector.process(request3)
    ssd_detector.update_response()
    log.info("---------------------")
    log.info("---------------------")


#############################
### START OF REAL TESTING ###
#############################

import boto3
import json
import os
from io import BytesIO

from utils import generate_fileobj_from_s3_folder

from multivitamin.applications.detectors import SSDDetector
from multivitamin.module_api.utils import load_idmap


S3_BUCKET_NAME = "vitamin-cv-test-data"
S3_EXPECTED_PREV_RESPONSES = "data/previous_responses"
S3_NET_DATA_FOLDER = "models/caffe/ssd/SportsNHLPlacementDetector-v1.2/"  # DO NOT TOUCH

LOCAL_NET_DATA_DIR = "/tmp/test_caffe_detector/net_data"


def test_init():
    global cc

    server_name = "TestDetector"
    version = "0.0.0"
    # net_data_dir = "/tmp/test_caffe_classifier/net_data"
    # prop_type = "label"
    # prop_id_map = {}
    # module_id_map
    if os.path.exists(LOCAL_NET_DATA_DIR):
        try:
            cc = CaffeClassifier(server_name, version, LOCAL_NET_DATA_DIR)
            return
        except:
            pass
    else:
        os.makedirs(LOCAL_NET_DATA_DIR)

    for key, net_data_bytes in generate_fileobj_from_s3_folder(
        S3_BUCKET_NAME, S3_NET_DATA_FOLDER
    ):
        filename = os.path.basename(key)
        print("{}/{}".format(LOCAL_NET_DATA_DIR, filename))
        with open("{}/{}".format(LOCAL_NET_DATA_DIR, filename), "wb") as file:
            file.write(net_data_bytes.getvalue())

    ssd = SSDDetector(server_name, version, LOCAL_NET_DATA_DIR)


def download_expected_response(path):
    s3 = boto3.client("s3")
    filelike = BytesIO()
    s3.download_fileobj(
        S3_BUCKET_NAME, S3_EXPECTED_PREV_RESPONSES + "/" + path, filelike
    )
    filelike.seek(0)
    expected_json = json.loads(filelike.read().decode())
    return expected_json


def test_consistency1():
    # Just an image
    pass


def test_consistency2():
    #  Just a video
    expected_json = download_expected_response("detection_doc.json")

    message = {"url": expected_json["media_annotation"]["url"]}
    module_map = {}
    module_map["NHLPlacementDetector"] = 60
    module_map["NHLLogoClassifier"] = 61
    module_map["NBAPlacementDetector"] = 69
    module_map["NBALogoClassifier"] = 70
    module_map["NBALeagueDetector"] = 71

    placement_id_map_file = LOCAL_NET_DATA_DIR + "/idmap.txt"
    placement_map = load_idmap(placement_id_map_file)

    ssd = SSDDetector(
        "NHLPlacementDetector",
        "0.0.1",
        LOCAL_NET_DATA_DIR,
        prop_type="placement",
        prop_id_map=placement_map,
        module_id_map=module_map,
    )
    ssd.process(message)
    ssd.update_response()
    j1 = json.dumps(expected_json, indent=2, sort_keys=True)
    j2 = json.dumps(ssd.avro_api.doc, indent=2, sort_keys=True)
    j2 = j2.replace(
        ssd.avro_api.doc["media_annotation"]["codes"][0]["date"],
        expected_json["media_annotation"]["codes"][0]["date"],
    )
    j2 = j2.replace(
        ssd.avro_api.doc["media_annotation"]["codes"][0]["id"],
        expected_json["media_annotation"]["codes"][0]["id"],
    )
    assert j1 == j2
    assert json.loads(j1) == json.loads(j2)



from abc import abstractmethod
import traceback
import pandas as pd
import glog as log

from .module import Module, Codes
from .utils import (pandas_query_matches_props,
                    batch_generator)
from ..media import MediaRetriever


MAX_PROBLEMATIC_FRAMES = 10
BATCH_SIZE = 1


class ImagesModule(Module):
    def __init__(
        self,
        server_name,
        version,
        prop_type=None,
        prop_id_map=None,
        module_id_map=None,
        batch_size=BATCH_SIZE,
    ):
        super().__init__(
            server_name=server_name,
            version=version,
            prop_type=prop_type,
            prop_id_map=prop_id_map,
            module_id_map=module_id_map,
        )
        self.batch_size = batch_size
        log.debug(f"Creating ImagesModule with batch_size: {batch_size}")

    def process(self, response):
        """Process the message, calls process_images(batch, tstamps, contours=None)
           which is implemented by the child module

        Returns:
            Response: response object
        """
        log.debug("Processing message")
        super().process(response)

        try:
            log.info(f"Loading media from {self.response.request}:"
                     f" {self.response.request.url}")
            self.media = MediaRetriever(self.response.request.url)
            self.frames_iterator = self.media.get_frames_iterator(
                self.response.request.sample_rate
            )
        except Exception as e:
            log.error(e)
            log.error(traceback.print_exc())
            self.code = Codes.ERROR_LOADING_MEDIA
            return self.update_and_return_response()

        self._update_w_h_in_response()

        if self.prev_pois and not self.response.has_frame_anns():
            log.warning("NO_PREV_REGIONS_OF_INTEREST, returning...")
            self.code = Codes.NO_PREV_REGIONS_OF_INTEREST
            return self.update_and_return_response()

        num_problematic_frames = 0
        for image_batch, tstamp_batch, prev_region_batch in batch_generator(
            self.preprocess_input(), self.batch_size
        ):
            if image_batch is None or tstamp_batch is None:
                continue
            try:
                self.process_images(image_batch,
                                    tstamp_batch,
                                    prev_region_batch)
            except ValueError as e:
                num_problematic_frames += 1
                log.warning("Problem processing frames")
                if num_problematic_frames >= MAX_PROBLEMATIC_FRAMES:
                    log.error(e)
                    self.code = Codes.ERROR_PROCESSING
                    return self.update_and_return_response()
        log.debug("Finished processing.")

        if self.prev_pois and self.prev_regions_of_interest_count == 0:
            log.warning("NO_PREV_REGIONS_OF_INTEREST, returning...")
            self.code = Codes.NO_PREV_REGIONS_OF_INTEREST
        return self.update_and_return_response()

    def preprocess_input(self):
        """Parses request for data

        Yields:
            frame: An image a time tstamp of a video or image
            tstamp: The timestamp associated with the frame
            region: The matching region dict
        """
        for i, (frame, tstamp) in enumerate(self.frames_iterator):
            if frame is None:
                log.warning("Invalid frame")
                continue

            if tstamp is None:
                log.warning("Invalid tstamp")
                continue

            self.tstamps_processed.append(tstamp)
            log.debug(f"tstamp: {tstamp}")
            if i % 100 == 0:
                log.info(f"Processing {self.response.request} inprogress:"
                         f" tstamp={tstamp}")

            if not self.prev_pois:
                yield frame, tstamp, None
            else:
                log.debug("Processing with previous response")
                log.debug(f"Querying on self.prev_pois: {self.prev_pois}")
                regions_that_match_props = []
                regions_at_tstamp = self.response.get_regions_from_tstamp(
                    tstamp
                )
                log.debug(f"Finding regions at tstamp: {tstamp}")
                if regions_at_tstamp is not None:
                    log.debug(f"len(regions_at_tstamp):"
                              f" {len(regions_at_tstamp)}")
                    for i_region in regions_at_tstamp:
                        if self._region_contains_props(i_region):
                            log.debug(f"region: {i_region} contains props"
                                      " of interest")
                            regions_that_match_props.append(i_region)
                            self.prev_regions_of_interest_count += 1

                for region in regions_that_match_props:
                    yield frame, tstamp, region

    @abstractmethod
    def process_images(self,
                       image_batch,
                       tstamp_batch,
                       prev_region_batch=None):
        """Abstract method to be implemented by child module"""
        pass

    def _update_w_h_in_response(self):
        (width, height) = self.media.get_w_h()
        log.debug(f"Setting in response w: {width} h: {height}")
        self.response.width = width
        self.response.height = height

    def _region_contains_props(self, region):
        """ Boolean to check if a region's props matches the defined
            previous properties of interest

        Args:
            props (list): list of properties for a region
        Returns:
            bool: if props match prev_pois query
        """
        props = region.get("props")
        if props is None:
            return False
        return pandas_query_matches_props(
            self.prev_pois_bool_exp,
            pd.DataFrame(props))



import os
import sys
import glog as log
import json
from pathlib import Path
from queue import Queue

from multivitamin.data import Request, Response
from multivitamin.apis.comm_api import CommAPI
from multivitamin.data.response.utils import get_current_date

FILEEXT = ".json"
DEFAULT_FILE = "response" + FILEEXT
INDENTATION = 2


class LocalAPI(CommAPI):
    def __init__(self, pulling_folder=None, pushing_folder=None, default_file=False):
        """LocalAPI is a CommAPI object that pulls queries from a local file
        and pushes responses to a file

        Args:
            pulling_folder (str): folder of jsons defining query, assumes *.json suffix
            pushing_folder (str): where to write responses
            default_file (bool): write to response.json and overwrite each time
        """
        self.pushing_folder = pushing_folder
        self.json_queue = Queue()
        self.default_file = default_file

        if pulling_folder:
            p = Path(pulling_folder)
            if not p.exists():
                raise ValueError(f"{pulling_folder} does not exist.")
            paths = list(p.glob("**/*{}".format(FILEEXT)))
            for path in paths:
                with path.open() as rf:
                    for row in rf:
                        self.json_queue.put(json.loads(row))

        if not os.path.exists(pushing_folder):
            os.makedirs(pushing_folder)

    def pull(self, n=1):
        """Pull a batch of messages to be processed

        Args:
            n (int): batch size
        
        Returns:
            list (Request): list of Request objects to process
        """
        log.info("Pulling local query jsons.")
        requests = []
        for _ in range(n):
            if self.json_queue.empty():
                log.info("Queue of jsons is empty. Exiting.")
                return [Request({"kill_flag": "true"})]
            m = self.json_queue.get()
            log.info(f"Appending request: {m}")
            requests.append(Request(m))
        return requests

    def push(self, responses):
        """Push a list of Response objects to be written to pushing_folder

        Args:
            request_apis (list[Response]): list of responses to be pushed/written to disk
        
        Returns:
            list[str]: list of response filenames written
        """
        if not isinstance(responses, list):
            responses = [responses]

        for response in responses:
            assert isinstance(response, Response)

        log.info(f"Pushing {len(responses)} items to folder: {self.pushing_folder}")
        outfns = []
        for res in responses:
            fn = self.get_fn(res)
            if not os.path.exists(os.path.dirname(fn)):
                os.makedirs(os.path.dirname(fn))

            log.info(f"Writing {fn}")
            if res.request.bin_encoding is True:
                with open(fn, "wb") as wf:
                    wf.write(res.to_bytes())
            else:
                with open(fn, "w") as wf:
                    wf.write(json.dumps(res.to_dict(), indent=INDENTATION))
            outfns.append(fn)
        return outfns

    def get_fn(self, response):
        """Create a fn from url string 

        Args:
            response (Response): response
        
        Returns:
            str: unique identifier
        """
        if self.default_file:
            return os.path.join(self.pushing_folder, DEFAULT_FILE)

        extension = "json"
        if response.request.bin_encoding is True:
            extension = "avro"
        media_url = response.url
        media_name = os.path.basename(media_url)
        return os.path.join(
            self.pushing_folder, f"{get_current_date()}", f"{media_name}.{extension}"
        )



import json
from abc import (ABC, abstractmethod)

import glog as log
from flask import jsonify

from multivitamin.data import Request, Response


class AiApplication(ABC):
    '''
        This class represent an AI Application. An AI Application is a list of modules to
        iterate over.
    '''
    models = None  # Where we keep the model when it's loaded

    @abstractmethod
    def get_modules(self):
        pass

    def predict(self, input):
        try:
            req = Request(input)
            response = Response(req)

            for module in self.get_modules():
                log.debug(f"Processing request for module: {module}")
                response = module.process(response)
                log.debug(f"response.to_dict(): {json.dumps(response.to_dict(), indent=2)}")

            if req.bin_encoding:
                return response.to_bytes()
            else:
                return jsonify(response.to_dict())
        except Exception:
            log.exception(f"Error processing request {input}")
        return jsonify({"error": "Could not process this request"})


import os
import json
from io import BytesIO

import boto3
import glog as log
import numpy as np
import traceback
from concurrent.futures import ThreadPoolExecutor
from PIL import Image


from multivitamin.module import PropertiesModule, Codes
from multivitamin.media import MediaRetriever
from multivitamin.data.response.dtypes import (
    VideoAnn,
    Property,
)


def get_contents_file_s3_key(url, sample_rate=None):
    contents_file_name = "contents"
    contents_file_ext = "json"
    rel_path_format = "{video_id}/{filename}.{ext}"
    video_id = os.path.basename(url).rsplit(".", 1)[0]
    video_id = "".join([e for e in video_id if e.isalnum() or e in ["/", "."]])
    if video_id is not None:
        video_id += ".{}fps".format(sample_rate)
    contents_file_key = rel_path_format.format(
        video_id=video_id, filename=contents_file_name, ext=contents_file_ext
    )
    contents_file_key = "".join(
        [e for e in contents_file_key if e.isalnum() or e in ["/", "."]]
    )
    return contents_file_key


class FrameExtractor(PropertiesModule):
    def __init__(
        self,
        server_name,
        version,
        sample_rate=1.0,
        s3_bucket=None,
        local_dir=None,
        module_id_map=None,
        n_threads=100,
    ):
        super().__init__(server_name, version, module_id_map=module_id_map)

        self.n_threads = n_threads
        self._sample_rate = sample_rate
        self._local_dir = local_dir
        self._s3_bucket = s3_bucket
        self._list_file = "contents"
        self._img_name_format = "{tstamp:010d}"
        self._rel_path_format = "{video_id}/{filename}.{ext}"
        self._image_rel_path_format = "{video_id}/frames/{filename}.{ext}"
        self._s3_url_format = "https://s3.amazonaws.com/{bucket}/{s3_key}"

        self._sql_db = ""

        self._s3_client = boto3.client("s3")

        self._encoding = "JPEG"
        self._content_type = "image/jpeg"
        self._s3_upload_args = {"ContentType": self._content_type}

    @staticmethod
    def _mklocaldirs(directory):
        if not os.path.exists(directory):
            os.makedirs(directory)

    def _write_frame_helper(self, data):
        try:
            self._write_frame(**data)
        except:
            log.error("Local Write Failed")

    def _write_frame(self, frame, tstamp, video_id):
        filename = self._img_name_format.format(tstamp=tstamp)
        relative_path = self._image_rel_path_format.format(
            video_id=video_id, filename=filename, ext=self._encoding
        )
        full_path = "{}/{}".format(self._local_dir, relative_path)
        full_path = "".join([e for e in full_path if e.isalnum() or e in ["/", "."]])
        im_filelike = self._convert_frame_to_filelike(frame)
        with open(full_path, "wb") as f:
            f.write(im_filelike.read())

    def _add_contents_to_local(self, contents):
        filelike = BytesIO()
        contents_json = {"original_url": self.video_url, "frames": []}
        for video_id, tstamp in contents:
            filename = self._img_name_format.format(tstamp=tstamp)
            relative_path = self._image_rel_path_format.format(
                video_id=video_id, filename=filename, ext=self._encoding
            )
            full_path = "{}/{}".format(self._local_dir, relative_path)
            full_path = "".join(
                [e for e in full_path if e.isalnum() or e in ["/", "."]]
            )
            # line = "{}\t{}\n".format(tstamp, full_path)
            # filelike.write(line.encode())
            contents_json["frames"].append((tstamp, full_path))
        filelike.write(json.dumps(contents_json, indent=2).encode())
        filelike.seek(0)

        with open("{}/{}".format(self._local_dir, self.contents_file_key), "wb") as f:
            f.write(filelike.read())

    def _convert_frame_to_filelike(self, frame):
        im = Image.fromarray(frame)
        im_filelike = BytesIO()
        im.save(im_filelike, format=self._encoding)
        im_filelike.seek(0)
        return im_filelike

    def _upload_frame_helper(self, data):
        try:
            self._upload_frame(**data)
        except:
            log.error("Failed to write Frame to S3")

    def _upload_frame(self, frame, tstamp, video_id):
        filename = self._img_name_format.format(tstamp=tstamp)
        im_filelike = self._convert_frame_to_filelike(frame)
        s3_key = self._image_rel_path_format.format(
            video_id=video_id, filename=filename, ext=self._encoding
        )
        s3_key = "".join([e for e in s3_key if e.isalnum() or e in ["/", "."]])
        result = self._s3_client.upload_fileobj(
            im_filelike, self._s3_bucket, s3_key, ExtraArgs=self._s3_upload_args
        )
        self._append_to_sql(video_id, tstamp, s3_key, result)

    def _append_to_sql(self, video_id, tstamp, s3_key, result):
        pass

    def _add_contents_to_s3(self, contents):
        filelike = BytesIO()
        contents_json = {"original_url": self.video_url, "frames": []}
        for video_id, tstamp in contents:
            filename = self._img_name_format.format(tstamp=tstamp)
            s3_key = self._image_rel_path_format.format(
                video_id=video_id, filename=filename, ext=self._encoding
            )
            s3_key = "".join([e for e in s3_key if e.isalnum() or e in ["/", "."]])
            im_url = self._s3_url_format.format(bucket=self._s3_bucket, s3_key=s3_key)
            # line = "{}\t{}\n".format(tstamp, im_url)
            # filelike.write(line.encode())
            contents_json["frames"].append((tstamp, im_url))
        filelike.write(json.dumps(contents_json, indent=2).encode())
        filelike.seek(0)

        result = self._s3_client.upload_fileobj(
            filelike,
            self._s3_bucket,
            self.contents_file_key,
            ExtraArgs={"ContentType": "application/json"},
        )
        return result

    def process_properties(self):
        self.last_tstamp = 0.0
        log.info("Processing")
        # filelike = self.media_api.download(return_filelike=True)

        # if filelike.getbuffer().nbytes == 0:
        #     self.code = "ERROR_NO_IMAGES_LOADED"

        # log.info('Getting hash')
        # video_hash = hashfileobject(filelike, hexdigest=True)
        self.video_url = self.response.request.url
        try:
            log.info(f"Loading media from url: {self.response.request.url}")
            self.med_ret = MediaRetriever(self.video_url)
        except Exception as e:
            log.error(e)
            log.error(traceback.print_exc())
            self.code = Codes.ERROR_LOADING_MEDIA
            return

        self.contents_file_key = get_contents_file_s3_key(self.video_url,
                                                          self._sample_rate)
        video_id = self.contents_file_key.split("/")[0]
        if self._local_dir is not None:
            self._mklocaldirs("{}/{}".format(self._local_dir, video_id))
            self._mklocaldirs("{}/{}/frames".format(self._local_dir, video_id))
            if os.path.exists("{}/{}".format(self._local_dir, self.contents_file_key)):
                log.info("Local Video already exists")

        try:
            self._s3_client.head_object(
                Bucket=self._s3_bucket, Key=self.contents_file_key
            )
            new_url = self._s3_url_format.format(
                bucket=self._s3_bucket, s3_key=self.contents_file_key
            )
            log.info("Video already exists")
            p = Property(
                server=self.name,
                ver=self.version,
                value=new_url,
                property_type="extraction",
                property_id=1,
            )
            track = VideoAnn(t1=0.0, t2=float(self.last_tstamp), props=[p])
            self.response.append_track(track)
            return
        except Exception:
            pass

        contents = []
        log.info("Getting frames")
        with ThreadPoolExecutor(max_workers=self.n_threads) as writer:
            for i, (frame, tstamp_secs) in enumerate(
                self.med_ret.get_frames_iterator(sample_rate=self._sample_rate)
            ):
                tstamp = int(tstamp_secs * 1000)
                # self._upload_frame(frame, tstamp, video_hash)
                if i % 100 == 0:
                    log.info("...tstamp: " + str(tstamp))
                log.debug("tstamp: " + str(tstamp))
                if frame is None:
                    continue
                frame = np.ascontiguousarray(frame[:, :, ::-1])  # RGB to BGR
                self.last_tstamp = tstamp
                data = {"frame": frame, "tstamp": tstamp, "video_id": video_id}
                contents.append((video_id, tstamp))

                if self._local_dir is not None:
                    writer.submit(self._write_frame_helper, data)

                if self._s3_bucket is not None:
                    writer.submit(self._upload_frame_helper, data)

        if self._s3_bucket is not None:
            _ = self._add_contents_to_s3(contents)

        if self._local_dir is not None:
            self._add_contents_to_local(contents)

        new_url = self._s3_url_format.format(
            bucket=self._s3_bucket, s3_key=self.contents_file_key
        )
        p = Property(
            server=self.name,
            ver=self.version,
            value=new_url,
            property_type="extraction",
            property_id=1,
        )
        track = VideoAnn(t1=0.0, t2=float(self.last_tstamp), props=[p])
        self.response.append_track(track)



# from multivitamin.data.response.data import MediaAnn, Point, ResponseInternal
# import json

# p = Point(1, 2)
# print(p)
# print(type(p))
# print(type(Point))
# assert isinstance(p, type(Point()))

# m = MediaAnn(w=44)
# print(type(m))
# print(m['w'])
# print(json.dumps(m.to_dict(), indent=2))
# x = m.to_dict()
# print(x)


# ri = ResponseInternal()

# ve = ri.to_dict()
# print(ve)
# maa = ResponseInternal.from_dict(ri.to_dict())
# # print(maa)
# # # x['z'] = 2.0
# # # print(x)
# # # # x['w'] = 33
# # import random

from collections import MutableMapping

from dataclasses import dataclass, field, asdict
from typing import List
from typeguard import typechecked
# from mashumaro import DataClassDictMixin


class DictLike(MutableMapping):
    """Base class used for the below dataclasses, so that each dataclass can 
    act like a dict with [] access
    """
    def __init__(self, *args, **kwargs):
        self.__dict__.update(*args, **kwargs)

    def __setitem__(self, key, value):
        self.__dict__[key] = value

    def __getitem__(self, key):
        return self.__dict__[key]

    def __delitem__(self, key):
        del self.__dict__[key]

    def __iter__(self):
        return iter(self.__dict__)

    def __len__(self):
        return len(self.__dict__)

    def __str__(self):
        return str(self.__dict__)

    def __repr__(self):
        return f"{super().__repr__()}, ({self.__dict__})"



@typechecked
@dataclass
class VideoAnn(DictLike):
    t1: float = 0.0
    t2: float = 0.0
    props: List[int] = field(default_factory=list)
    regions: List[float] = field(default_factory=list)
    region_ids: List[str] = field(default_factory=list)

v = VideoAnn()
print(v)
dv = asdict(v)
print(dv)
xe = VideoAnn(**dv)
print(xe)


import sys
import os
import glog as log

sys.path.append(os.path.abspath("../examples/tf-detector/"))
import json

from tf_detector import TFDetector
import boto3
import zipfile

request1 = '{"url":"https://www.astrosafari.com/download/file.php?id=1608&sid=79f086b975b702945ca90b0ac3bd0202","dst_url":"https://vertex-api-v3.gumgum.com/v1/module-responses?featureId=5&moduleId=1&imageUrl\
=https%3A%2F%2Fwww.astrosafari.com%2Fdownload%2Ffile.php%3Fid%3D1608%26sid%3D79f086b975b702945ca90b0ac3bd0202&order=2&async=true","prev_response":"AAAAAEAGMi4wJjIwMTgtMTAtMTZUMTc6MjM6MjAAAAAAAAAAAAAAAhxOU0\
ZXQ2xhc3NpZmllcgAKNC4wLjEOU1VDQ0VTUwxndW1ndW0cMjAxODEwMTYxNzIzMjAObWFjaGluZQAAOk5TRldDbGFzc2lmaWVyXzIwMTgxMDE2MTcyMzIwALwBaHR0cHM6Ly93d3cuYXN0cm9zYWZhcmkuY29tL2Rvd25sb2FkL2ZpbGUucGhwP2lkPTE2MDgmYW1wO3NpZD0\
3OWYwODZiOTc1YjcwMjk0NWNhOTBiMGFjM2JkMDIwMrwBaHR0cHM6Ly93d3cuYXN0cm9zYWZhcmkuY29tL2Rvd25sb2FkL2ZpbGUucGhwP2lkPTE2MDgmYW1wO3NpZD03OWYwODZiOTc1YjcwMjk0NWNhOTBiMGFjM2JkMDIwMgCACsAHAgAAAAAChgEwLjAwMDBfKDAuMDAw\
MCwwLjAwMDApKDEuMDAwMCwwLjAwMDApKDEuMDAwMCwxLjAwMDApKDAuMDAwMCwxLjAwMDApAAgAAAAAAAAAAAAAgD8AAAAAAACAPwAAgD8AAAAAAACAPwACEgIcTlNGV0NsYXNzaWZpZXIACjQuMC4xDGd1bWd1bQxzYWZldHkIc2FmZQAc1nk/CtcjPAAAgD8AOk5TRldDb\
GFzc2lmaWVyXzIwMTgxMDE2MTcyMzIwAAAAAAIAAAAAAAAAAAISAhxOU0ZXQ2xhc3NpZmllchJoaXN0b2dyYW0KNC4wLjEMZ3VtZ3VtDHNhZmV0eQhzYWZlABzWeT8K1yM8AACAPwA6TlNGV0NsYXNzaWZpZXJfMjAxODEwMTYxNzIzMjAAAAAAAAA=","bin_encoding":"\
true","bin_decoding":"true"}'
request2 = '{"url":"http://www.soundbridgedentalarts.com/wp-content/uploads/X01014_1.jpg","prev_response":"","bin_encoding":"false","bin_decoding":"false"}'
request3 = '{"url":"http://pbs.twimg.com/media/Dpq40egWkAAtocH.jpg:large","dst_url":"https://vertex-api-v3.gumgum.com/v1/module-responses?featureId=26&moduleId=1004&imageUrl=http%3A%2F%2Fpbs.twimg.com%2Fmedia%2FDpq40egWkAAtocH.jpg%3Alarge&order=1&async=true","prev_response":"","bin_encoding":"true","bin_decoding":"true"}'

message = ""


def test_object_detector():
    log.info("Testing TF detector.")
    tmp_filepath = "/tmp/net_data.zip"

    log.info("Downloading net_data")
    s3_client = boto3.client("s3")
    log.info("Downloading model.")
    s3_client.download_file("vitamincv-data", "tf-detector/net_data.zip", tmp_filepath)
    log.info("Model downloaded.")

    with open(tmp_filepath, "rb") as f:
        log.info("Unzipping it.")
        z = zipfile.ZipFile(f)
        for name in z.namelist():
            print("    Extracting file", name)
            z.extract(name, "/tmp/")
    log.info("Unzipped.")
    tf_detector = TFDetector(
        server_name="TFDetector", version="1.0", net_data_dir="/tmp/net_data/"
    )
    log.info("1---------------------")
    tf_detector.process(request1)
    log.info("2---------------------")
    tf_detector.process(request2)
    log.info("3---------------------")
    tf_detector.process(request3)
    tf_detector.update_response()
    log.info("---------------------")
    log.info("---------------------")



import os
import pytest
import json
import csv
import boto3

from vitamincv.avro_api.avro_api import AvroAPI
from vitamincv.avro_api.avro_io import AvroIO
from vitamincv.avro_api.cv_schema_factory import *
import zipfile2

import glog as log

avro_str = "avro.json"
avro_bin = "avro.bin"

image_ann = create_image_ann(
    t=5.0,
    regions=[create_region(props=[create_prop(server="test_dummy", value="exists")])],
)

annotation_task_1 = create_annotation_task(
    id="task_dummy_1",
    tstamps=[0],
    labels=[
        create_eligible_prop(
            server="LeagueClassifier",
            property_type="league",
            value="NBA",
            confidence_min=0.9,
            father_properties=[],
        )
    ],
    tags=["dummy_1"],
)


def test_get_detections_from_frame_anns():
    s3_client = boto3.client("s3")
    log.info("Downloading json.")
    tmp_filepath = (
        "/tmp/NHL_GAME_VIDEO_WPGOTT_M2_HOME_20180402_1520698435976.t.mp4.json"
    )
    s3_client.download_file(
        "cvapis-data",
        "jsons/NHL_GAME_VIDEO_WPGOTT_M2_HOME_20180402_1520698435976.t.mp4.json",
        tmp_filepath,
    )
    log.info("JSON downloaded.")
    x = AvroAPI(AvroIO.read_json(tmp_filepath))
    print(json.dumps(x.get_detections_from_frame_anns(), indent=2))


def test_create_avro_str():
    avro_api = AvroAPI()
    avro_api.append_image_ann(image_ann)
    avro_api.set_url("blah")
    AvroIO.write_json(avro_api.get_response(), avro_str, indent=2)
    assert True


def test_create_avro_bin():
    avro_api = AvroAPI()
    avro_api.append_image_ann(image_ann)
    avro_api.set_url("blah")
    avro_io = AvroIO()
    response = avro_api.get_response()
    avro_io.write(response, avro_bin)
    flag = AvroIO.is_valid_avro_doc_static(response, avro_io.get_schema())
    assert flag


def test_create_avro_bin_w_registry():
    avro_api = AvroAPI()
    avro_api.append_image_ann(image_ann)
    avro_api.set_url("blah")
    avro_io = AvroIO()
    avro_io.write(avro_api.get_response(), avro_bin)
    assert True


"""
Example function to test reading json from S3
Json located at https://s3.amazonaws.com/cvapis-data/jsons/Threatpipeline_response0_C.json in SX/AI account
"""


def test_read_avro_json():
    s3_client = boto3.client("s3")
    log.info("Downloading json.")
    tmp_filepath = "/tmp/Threatpipeline_response0_C.json"
    s3_client.download_file(
        "cvapis-data", "jsons/Threatpipeline_response0_C.json", tmp_filepath
    )
    log.info("JSON downloaded.")
    avro_io = AvroIO()
    avro_str = avro_io.read_json(tmp_filepath)
    # log.info("avro_str: " + str(avro_str))
    assert len(avro_str) > 0
    x = AvroAPI(avro_str)


"""
Example function to get detections from json (parsing)
Json located at https://s3.amazonaws.com/cvapis-data/jsons/Threatpipeline_response0_C.json in SX/AI account
"""


def test_query_properties():
    s3_client = boto3.client("s3")
    log.info("Downloading json.")
    tmp_filepath = "/tmp/Threatpipeline_response0_C.json"
    s3_client.download_file(
        "cvapis-data", "jsons/Threatpipeline_response0_C.json", tmp_filepath
    )
    log.info("JSON downloaded.")
    avro_io = AvroIO()
    avro_str = avro_io.read_json(tmp_filepath)
    # log.info("avro_str: " + str(avro_str))
    assert len(avro_str) > 0
    x = AvroAPI(avro_str)
    p = {}
    p["server"] = "ThreatClassifier"
    props = [p]
    dets = x.get_detections_from_props(props)
    assert len(dets) == 1
    log.info("dets: " + str(dets))


"""
Example function to read folder (zipped) of jsons from S3, and get detections from
them based on a filter property; here, server_name
Zipped folder at https://s3.amazonaws.com/cvapis-data/jsons/threat_eg_jsons.zip in SX/AI account
"""


def test_parse_jsons_from_folder():
    s3_client = boto3.client("s3")
    log.info("Downloading zipped jsons.")
    name = "threat_eg_jsons"
    tmp_filepath = "/tmp/" + name + ".zip"
    s3_client.download_file("cvapis-data", "jsons/" + name + ".zip", tmp_filepath)
    log.info("Zipped jsons downloaded.")
    with open(tmp_filepath, "rb") as f:
        log.info("Unzipping it.")
        z = zipfile2.ZipFile(f)
        for name_local in z.namelist():
            print("    Extracting file", name_local)
            z.extract(name_local, "/tmp/")

    jsons_path = "/tmp/" + name
    log.info("Unzipped to" + jsons_path)
    log.info("Parrsing jsons in " + jsons_path)
    all_jsons = os.listdir(jsons_path)
    dets_per_media = {}
    server_name = "ThreatClassifier"
    for ind, single_json_rel in enumerate(all_jsons):
        single_json = jsons_path + "/" + single_json_rel
        log.info("Parsing: " + single_json)
        avro_io = AvroIO()
        x = AvroAPI(avro_io.read_json(single_json))
        p = {}
        p["server"] = server_name
        props = [p]
        dets = x.get_detections_from_props(props)
        url = x.get_url()
        dets_per_media[url] = dets
        log.info("\n\n\n")
        log.info("url: " + url)
        log.info("dets: " + str(dets))


def test_read_avro_bin():
    assert os.path.exists(avro_bin)
    avro_io = AvroIO()
    avro_str = avro_io.decode_file(avro_bin)
    assert avro_str


def test_append_annotation_task():
    avro_api = AvroAPI()
    avro_api.append_image_ann(image_ann)
    avro_api.set_url("blah")
    avro_api.append_annotation_task(annotation_task_1)
    avro_io = AvroIO()
    response = avro_api.get_response()
    #    print(json.dumps(response))
    #   avro_io.encode_to_file(response, avro_bin)
    flag = AvroIO.is_valid_avro_doc_static(response, avro_io.get_schema())
    assert flag
    AvroIO.write_json(response, annotation_task_1["id"] + ".json", 2)


def test_append_trackssummary_from_csv():
    jsons_folder = "../cvapis/data/jsons/"
    # input_json_filepath=jsons_folder + 'Tampa-Bay-Lightning-vs-Montreal-Canadians_03102018_112_108.json'
    # input_csv_filepath=jsons_folder + 'Tampa-Bay-Lightning-vs-Montreal-Canadians_03102018_112_108_GT.csv'
    input_json_filepath = (
        jsons_folder + "NHL_GAME_VIDEO_ANACHI_M2_HOME_20182016_1518124767584.json"
    )
    input_csv_filepath = (
        jsons_folder + "NHL_GAME_VIDEO_ANACHI_M2_HOME_20182016_1518124767584.csv"
    )
    assert os.path.exists(input_json_filepath)
    assert os.path.exists(input_csv_filepath)
    filename = os.path.basename(input_json_filepath)
    filename = os.path.splitext(filename)[0]
    output_json_filepath = jsons_folder + filename + "_With_Human_Tracks.json"
    # we load the input json
    avro_api = AvroAPI(AvroIO.read_json(input_json_filepath))
    # we parse the csv
    cois = ["Sponsor", "Placement", "Start Timestamp", "Stop Timestamp"]
    isponsor = -1
    iplacement = -1
    it1 = -1
    it2 = -1
    with open(input_csv_filepath) as csvfile:
        readCSV = csv.reader(csvfile, delimiter=",")
        titles = []
        placements = []
        sponsors = []
        t1s = []
        t2s = []

        for i, row in enumerate(readCSV):
            if i == 0:
                titles = row
                log.info("titles: " + str(titles))
                log.info("Getting indexes of interest for columns named:  " + str(cois))
                isponsor = titles.index(cois[0])
                iplacement = titles.index(cois[1])
                it1 = titles.index(cois[2])
                it2 = titles.index(cois[3])
                log.info(
                    "indexes: "
                    + str(isponsor)
                    + ", "
                    + str(iplacement)
                    + ", "
                    + str(it1)
                    + ", "
                    + str(it2)
                )
            else:
                placements.append(row[iplacement])
                sponsors.append(row[isponsor])
                t1aux = row[it1]
                t2aux = row[it2]
                if t1aux.find(":") == 1:
                    c1 = t1aux.split(":")
                    c2 = t2aux.split(":")
                    t1 = int(c1[0]) * 3600 + int(c1[1]) * 60 + int(c1[2])
                    t2 = int(c2[0]) * 3600 + int(c2[1]) * 60 + int(c2[2])
                else:  # we assume its seconds
                    t1 = int(t1aux)
                    t2 = int(t2aux)
                t1s.append(t1)
                t2s.append(t2)

        log.info("placements[0:9]: " + str(placements[0:9]))
        log.info("sponsors[0:9]: " + str(sponsors[0:9]))
        log.info("t1s[0:9]: " + str(t1s[0:9]))
        log.info("t2s[0:9]: " + str(t2s[0:9]))
        log.info("placements[-9:]: " + str(placements[-9:]))
        log.info("sponsors[-9:]: " + str(sponsors[-9:]))
        log.info("t1s[-9:]: " + str(t1s[-9:]))
        log.info("t2s[-9:]: " + str(t2s[-9:]))

    for t1, t2, placement, sponsor in zip(t1s, t2s, placements, sponsors):
        # we create the placecement property
        p1 = create_prop(
            confidence=1,
            ver="1.0",
            server="HAM",
            property_type="placement",
            value=placement,
        )
        # we create the sponsor property
        p2 = create_prop(
            confidence=1,
            ver="1.0",
            server="HAM",
            property_type="sponsor",
            value=sponsor,
        )
        ps = [p1, p2]
        # We create the track
        track = create_video_ann(t1=t1, t2=t2, props=ps)
        avro_api.append_track_to_tracks_summary(track)
    AvroIO.write_json(avro_api.get_response(), output_json_filepath, indent=2)
    assert True
    print("done")


# test_append_trackssummary_from_csv()
# def test_avro_created_w_cpp():
#     pass
#
# def test_valid_json_schema():
#     #assert(os.path.exists(avro_bin))
#     assert(os.path.exists(avro_str))
#     #avro_io = AvroIO()
#     #json_str_from_bin = avro_io.decode_file(avro_bin)
#     json_str = AvroIO.read_json(avro_str)
#     #assert(json_str == json_str_from_bin)
#     assert(AvroIO.validate_schema(json_str, schema_str) is True)
#
# def test_invalid_json_schema():
#     json_str = json.load(open("face_media_ann.json"))
#     assert(AvroIO.validate_schema(json_str, schema_str) is False)


#     media_ann = json.loads(open("test.json", "r").read())
#     xavro = avro_api.AvroAPI()
#     xavro.set_media_ann(media_ann)
#
#     p = {}
#     p["company"] = "gumgum"
#     p["server"] = "FaceDetector"
#     p["ver"] = ""
#     p["property_type"] = "label"
#     p["value"] = "face"
#     regions = xavro.get_regions_from_prop(p)
#     print(regions)
#     print(len(regions))
# #check that i can read a string json



import os
import sys
import json

from vitamincv.avro_api.avro_io import AvroIO

local_json = "data/placement.json"

if not os.path.exists(local_json):
    print("Execute pytest test_avro_io.py from within the tests/ folder")
    sys.exit(1)

temp_file = "tmp.avro"


def test_local_encode():
    avroio = AvroIO()
    assert avroio.write(AvroIO.read_json(local_json), temp_file)


def test_local_decode():
    avroio = AvroIO()
    res = avroio.decode_file(temp_file)
    assert True


def test_remote_encode():
    avroio = AvroIO(use_schema_registry=True)
    assert avroio.write(AvroIO.read_json(local_json), temp_file)


def test_remote_decode():
    avroio = AvroIO(use_schema_registry=True)
    doc = avroio.decode_file(temp_file)
    print(json.dumps(doc, indent=2))
    assert True


def test_is_valid_avro_doc():
    avroio = AvroIO()
    doc = avroio.decode_file(temp_file)
    assert avroio.is_valid_avro_doc(doc)



from enum import Enum


class Codes(Enum):
    """An Enum type for return codes"""

    SUCCESS = 0
    NO_PREV_REGIONS_OF_INTEREST = 1
    ERROR_LOADING_MEDIA = 2
    ERROR_PROCESSING = 3
    ERROR_INCORRECT_SAMPLERATE = 4
    ERROR_INCORRECT_PREVIOUS_RESPONSE = 5
    ERROR_SERIALIZING_RESPONSE = 6



import os
import argparse
from glob import glob
import boto3

sqs = boto3.resource('sqs')
a = argparse.ArgumentParser()
a.add_argument("--queue_name")
a.add_argument("--requests_dir")
args = a.parse_args()

queue = sqs.get_queue_by_name(QueueName=args.queue_name)
for req in glob(os.path.join(args.requests_dir, "*")):
    with open(req) as rf:
        resp = queue.send_message(MessageBody=rf.read().strip())
        print(resp)



import os
import sys
import glog as log
import json
import pkg_resources
import tempfile
import struct
import base64
import traceback

import avro.schema
from avro.datafile import DataFileReader, DataFileWriter
from avro.io import DatumReader, DatumWriter, BinaryDecoder, BinaryEncoder
from confluent_kafka.avro.cached_schema_registry_client import CachedSchemaRegistryClient
from confluent_kafka.avro.serializer.message_serializer import (
    MessageSerializer,
    ContextStringIO,
    MAGIC_BYTE,
)
from confluent_kafka.avro.serializer import SerializerError

from multivitamin.data.response import config


class AvroIO:
    def __init__(self, schema_registry_url=None):
        """Public interface for Avro IO functionality
        
        Args:
            use_schema_registry (bool): flag to use schema registry via client ID and registry URL
            use_base64 (bool): encoding binary to base64
        """
        self.impl = None
        self.use_base64 = False
        if schema_registry_url:
            log.info(f"schema_registry_url: {schema_registry_url}")
            self.impl = _AvroIORegistry(schema_registry_url)
        else:
            log.warning("registry_url is None, using local schema and serializing w/o magic byte")
            self.impl = _AvroIOLocal()

    def get_schema(self):
        """Return schema 

        Returns:
            avro.schema.RecordSchema: schema
        """
        return self.impl.schema

    def get_schema_str(self):
        """Return schema as str

        Returns:
            str: schema as str
        """
        return str(self.impl.schema).replace("\\", "")

    def decode_binary_file(self, file_path):
        """Decode an Avro Binary using the schema

        Args:
            file_path (str) : avro binary file

        Returns:
            dict: avro document
        """
        if not os.path.exists(file_path):
            log.error("Missing: {}".format(file_path))
            raise FileNotFoundError("Missing: {}".format(file_path))
        log.info("Decoding file: {}".format(file_path))
        return self.decode(open(file_path, "rb").read())

    def decode(self, bytes, use_base64=False, binary_flag=True):
        """Decode an Avro Binary using the CV schema from bytes
        
        Args:
            bytes
            use_base64
            binary_flag
        
        Returns:
            dict:
        """
        if use_base64:
            bytes = base64.b64decode(bytes)
        if binary_flag:
            return self.impl.decode(bytes)
        else:
            return json.loads(str(bytes))

    def write(self, doc, file, serialize=True, indent=None):
        """Write Avro doc 

        Args:
            doc (dict): dict of the avro document
            file_path (str): avro binary file output
            serialize (bool): whether to serialize avro doc to a binary file
            indent (int): if serialize=False, write json with indentation=indent
        
        Returns:
            bool: True if successfully wrote file
        """
        if serialize:
            try:
                bytes = self.encode(doc)
                with open(file, "wb") as wf:
                    wf.write(bytes)
            except avro.io.AvroTypeException:
                log.error("avro.io.AvroTypeException: the datum is not an example of the schema")
                return False
            log.info("Encoded doc to file: {}".format(file))
        else:
            if not self.is_valid_avro_doc(doc):
                log.error("datum is not an example of schema")
                return False
            with open(file, "w") as wf:
                json.dump(doc, wf, indent=indent)
        return True

    def encode(self, doc, use_base64=False):
        """Encode an avro doc to bytes
        """
        bytes = self.impl.encode(doc)
        if use_base64:
            log.info(f"use_base64={use_base64}")
            bytes = base64.b64encode(bytes)
        return bytes

    def is_valid_avro_doc(self, doc):
        """Boolean test to validate json against a schema

        Args:
            doc (dict): avro doc as a dict

        Returns:
            boolean: True if json is an example of schema
        """
        try:
            writer = DataFileWriter(tempfile.TemporaryFile(), DatumWriter(), self.impl.schema)
            writer.append(doc)
            writer.close()
        except:
            return False
        return True

    @staticmethod
    def is_valid_avro_doc_static(doc, schema):
        """Boolean test to validate json against a schema

        Args:
            doc (dict): avro doc as a dict
            schema (str or dict): schema as a string or dict

        Returns:
            boolean: True if json is an example of schema
        """
        if isinstance(schema, str):
            avro_schema = avro.schema.Parse(schema)
        else:
            avro_schema = schema
        try:
            writer = DataFileWriter(tempfile.TemporaryFile(), DatumWriter(), avro_schema)
            writer.append(doc)
            writer.close()
        except:
            return False
        return True


class _AvroIOLocal:
    def __init__(self):
        """Private implementation class for Avro IO of local files"""
        local_schema_file = pkg_resources.resource_filename(
            "multivitamin.data.response", config.SCHEMA_FILE
        )
        log.debug("Using local schema file {}".format(local_schema_file))
        if not os.path.exists(local_schema_file):
            raise FileNotFoundError("Schema file not found")
        self.schema = avro.schema.Parse(open(local_schema_file).read())

    def decode(self, bytes):
        if len(bytes) <= 5:
            raise SerializerError("Message is too small to decode")
        with ContextStringIO(bytes) as payload:
            magic, schema_id = struct.unpack(">bI", payload.read(5))
            if magic != MAGIC_BYTE:
                raise SerializerError("message does not start with magic byte")
            curr_pos = payload.tell()
            avro_reader = avro.io.DatumReader(self.schema)

            def decoder(p):
                bin_decoder = avro.io.BinaryDecoder(p)
                return avro_reader.read(bin_decoder)

            return decoder(payload)

    def encode(self, record):
        with ContextStringIO() as outf:
            outf.write(struct.pack("b", MAGIC_BYTE))
            outf.write(struct.pack(">I", config.SCHEMA_ID))
            encoder = avro.io.BinaryEncoder(outf)
            writer = avro.io.DatumWriter(self.schema)
            writer.write(record, encoder)
            return outf.getvalue()


class _AvroIORegistry:
    def __init__(self, schema_registry_url):
        """Private implementation class for Avro IO using the registry"""
        log.info(f"Using registry with schema_url/id {schema_registry_url}/{config.SCHEMA_ID}")
        try:
            self.client = CachedSchemaRegistryClient(url=schema_registry_url)
            self.schema = self.client.get_by_id(config.SCHEMA_ID)
            self.serializer = MessageSerializer(self.client)
        except:
            raise ValueError("Client id or schema id not found")

    def decode(self, bytes):
        return self.serializer.decode_message(bytes)

    def encode(self, record):
        return self.serializer.encode_record_with_schema_id(config.SCHEMA_ID, record)



import numpy as np
import pytest
import random
import glog as log

from multivitamin.media import OpenCVMediaRetriever

VIDEO_URL = "https://s3.amazonaws.com/video-ann-testing/NHL_GAME_VIDEO_NJDMTL_M2_NATIONAL_20180401_1520698069177.t.mp4"
IMAGE_URL = "https://upload.wikimedia.org/wikipedia/commons/0/0b/Cat_poster_1.jpg"

VIDEO_CODEC_PROB_1 = "https://s3.amazonaws.com/video-ann/538_Pelicans+vs+Thunder+11%3A5-fhj713lbrhi.30-31.mp4"
VIDEO_CODEC_PROB_2 = "https://s3.amazonaws.com/gumgum-sports-analyst-data/media-files/Replay%20Video%20Capture_2018-11-16_11.52.51-2816an1tb0v.mp4"
VIDEO_CODEC_PROB_3 = "https://s3.amazonaws.com/gumgum-sports-analyst-data/media-files/1%3A3%20Houston%20Rockets%20at%20Golden%20State%20Warriors-6tgm4my1dr6.mp4"

VIDEO_URLS = [VIDEO_URL, VIDEO_CODEC_PROB_1, VIDEO_CODEC_PROB_2, VIDEO_CODEC_PROB_3]


def test_bad_url():
    with pytest.raises(FileNotFoundError):
        med_ret = OpenCVMediaRetriever("blah")


def test_not_a_media_url():
    with pytest.raises(ValueError):
        med_ret = OpenCVMediaRetriever("http://www.google.com")


def test_not_a_valid_filepath():
    with pytest.raises(FileNotFoundError):
        med_ret = OpenCVMediaRetriever("file://fhekslf")


@pytest.mark.parametrize("video_url", VIDEO_URLS)
def test_attributes(video_url):
    mr = OpenCVMediaRetriever(video_url)

    assert(mr.get_fps() > 0)
    assert(mr.get_num_frames() > 0)
    assert(mr.get_length() > 0)
    w, h = mr.get_w_h()
    assert(w > h > 0)  # Assume all videos are wider than tall.


def test_image():
    mr = OpenCVMediaRetriever(IMAGE_URL)
    assert(mr.image is not None)
    assert(mr.is_image)


def test_download():
    mr = OpenCVMediaRetriever(IMAGE_URL)
    filelike_obj = mr.download(return_filelike=True)
    assert filelike_obj
    assert len(filelike_obj.read()) > 0


# @pytest.mark.parametrize("video_url", VIDEO_URLS)
# def test_get_frame(video_url):
#     efficient_mr, fast_mr = create_media_retrievers(video_url)
#     assert efficient_mr.get_length() == fast_mr.get_length()
#     length = efficient_mr.get_length()
#     random_tstamp = length * random.random()
#     im1 = efficient_mr.get_frame(random_tstamp)
#     im2 = fast_mr.get_frame(random_tstamp)
#     assert np.array_equal(im1, im2)
#
#
# @pytest.mark.parametrize("video_url", VIDEO_URLS)
# def test_frames_iterator(video_url):
#     efficient_mr, fast_mr = create_media_retrievers(video_url)
#     assert efficient_mr.get_length() == fast_mr.get_length()
#     length = efficient_mr.get_length()
#     random_tstamp1 = length * random.random()
#     random_tstamp2 = length * random.random()
#     sample_rate1 = 0.1 * efficient_mr.get_fps()
#     sample_rate2 = 10 * efficient_mr.get_fps()
#     _run_frames_iterator(
#         efficient_mr,
#         fast_mr,
#         sample_rate1,
#         min(random_tstamp1, random_tstamp2),
#         max(random_tstamp1, random_tstamp2),
#     )
#     _run_frames_iterator(
#         efficient_mr,
#         fast_mr,
#         sample_rate2,
#         min(random_tstamp1, random_tstamp2),
#         max(random_tstamp1, random_tstamp2),
#     )
#
#
# def _run_frames_iterator(efficient_mr, fast_mr, sample_rate, start, stop):
#     efficient_iterator = efficient_mr.get_frames_iterator(
#         sample_rate=sample_rate, start_tstamp=start, end_tstamp=stop
#     )
#     fast_iterator = fast_mr.get_frames_iterator(
#         sample_rate=sample_rate, start_tstamp=start, end_tstamp=stop
#     )
#     stopped1 = False
#     stopped2 = False
#     for idx in range(100):
#         log.info(idx)
#         try:
#             im1, t1 = next(efficient_iterator)
#         except StopIteration:
#             log.info("Efficient Video Ended Early")
#             stopped1 = True
#
#         try:
#             im2, t2 = next(fast_iterator)
#         except StopIteration:
#             log.info("Fast Video Ended Early")
#             stopped2 = True
#
#         assert stopped1 == stopped2
#         assert np.array_equal(im1, im2)
#         assert t1 == t2
#
#
# @pytest.mark.parametrize("video_url", VIDEO_URLS)
# def test_consistency_between_get_frame_and_frames_iterator(video_url):
#     efficient_mr, fast_mr = create_media_retrievers(video_url)
#     assert efficient_mr.get_length() == fast_mr.get_length()
#     length = efficient_mr.get_length()
#     random_tstamp = length * random.random()
#     efficient_iterator = efficient_mr.get_frames_iterator(start_tstamp=random_tstamp)
#     fast_iterator = fast_mr.get_frames_iterator(start_tstamp=random_tstamp)
#     im1_iter, t1_iter = next(efficient_iterator)
#     im2_iter, t2_iter = next(fast_iterator)
#     im1 = efficient_mr.get_frame(t1_iter)
#     im2 = fast_mr.get_frame(t2_iter)
#     assert np.array_equal(im1_iter, im1)
#     assert np.array_equal(im2_iter, im2)
#     assert np.array_equal(im1, im2)



from abc import ABC, abstractmethod


class CommAPI(ABC):
    """Abstract base class to define an interface of `push()` and `pull()`"""
    @abstractmethod
    def pull(self):
        pass

    @abstractmethod
    def push(self):
        pass



import json
import traceback

import glog as log
from dataclasses import asdict

from multivitamin.data import Request
from multivitamin.data.response.io import AvroIO
from multivitamin.data.response.dtypes import (
    ResponseInternal,
    Region,
    VideoAnn,
    ImageAnn,
    Footprint,
)
from multivitamin.data.response.utils import round_float
from multivitamin.media.file_retriever import FileRetriever


class Response():
    def __init__(self, response_input=None, schema_registry_url=None):
        """ Class for a Response object
        
        2 cases for construction:

                1) constructed from a request with a previous Response
                2) constructing from a dictionary

        Args:
            input (Any): previous Response or dict
            schema_registry_url (str): whether to use schema registry when serializing to bytes
        """
        self._request = None
        self._response_internal = None
        self._schema_registry_url = schema_registry_url
        self._tstamp2frameannsidx = {}

        if isinstance(response_input, Request):
            self._request = response_input
            self._init_from_request()
        elif isinstance(response_input, dict):
            io = AvroIO(schema_registry_url)
            if not io.is_valid_avro_doc(response_input):
                raise ValueError("Input dict is incompatible with avro schema")
            log.debug("Input dict is compatible with avro schema")

            # unpack dictionary values into kwargs using ** operator
            try:
                self._response_internal = ResponseInternal(**response_input)
            except Exception as e:
                log.error("error unpacking prev_response_dict")
                log.error(traceback.format_exc())
        else:
            log.debug("Initializing empty response")
            self._response_internal = ResponseInternal()

        self._init_tstamp2frameannsidx()

    def to_dict(self):
        """Getter for response in the form of a dict

        Returns:
            dict: response as dict
        """
        log.debug("Returning response as dictionary")
        if self._request is not None:
            if self._request.bin_encoding is True:
                log.warning("self._request.bin_encoding is True but returning dictionary")
        return asdict(self._response_internal)

    def to_bytes(self, base64=False):
        """Getter for response in the form of bytes or base64 str

        Args:
            base64 (bool): flag for converting to base64 encoding
        
        Returns:
            bytes: response as bytes
        """
        if self._request is not None:
            base64 = self._request.base64_encoding
            log.info("Using self._request to check base64_encoding flag")
        log.debug(f"base64 encoding: {base64}")
        log.debug("Returning response as binary")
        try:
            io = AvroIO(self._schema_registry_url)
            return io.encode(asdict(self._response_internal), base64)
        except Exception:
            log.error("Error serializing response")
            # what to do here?
            raise Exception("Error serializing response")

    @property
    def data(self):
        """Convenience getter to return either dict or bytes depending on request

        Returns:
            Any: dict or bytes
        """
        if self._request is None:
            return self.to_dict()

        if self._request.bin_encoding is False:
            log.debug("Returning response as dictionary")
            return self.to_dict()

        return self.to_bytes(self._request.base64_encoding)

    @property
    def tracks(self):
        """Getter for returning tracks

        Returns:
            List[VideoAnn]: tracks
        """
        return self._response_internal["media_annotation"]["tracks_summary"]

    @tracks.setter
    def tracks(self, tracks):
        """Setter for tracks

        Args:
            tracks (List[VideoAnn]): tracks
        """
        self._response_internal["media_annotation"]["tracks_summary"] = tracks

    @property
    def frame_anns(self):
        """Getter for frames annotations

        Returns:
            List[ImageAnn]: frames ann
        """
        return self._response_internal["media_annotation"]["frames_annotation"]

    def has_frame_anns(self):
        """Bool check on whether there are frame anns

        Returns:
            bool: flag for existence of frame anns
        """
        return len(self.frame_anns) > 0

    def get_regions_from_tstamp(self, t):
        """Get regions for a timestamp

        Args:
            t (float): tstamp
        
        Returns:
            List[Region]: regions
        """
        assert isinstance(t, float)
        if t not in self._tstamp2frameannsidx:
            return None
        frame_anns_idx = self._tstamp2frameannsidx[t]
        return self._response_internal["media_annotation"]["frames_annotation"][frame_anns_idx]["regions"]

    @property
    def request(self):
        return self._request

    @property
    def url(self):
        return self._response_internal["media_annotation"]["url"]

    @url.setter
    def url(self, url):
        assert isinstance(url, str)
        self._response_internal["media_annotation"]["url"] = url

    @property
    def url_original(self):
        return self._response_internal["media_annotation"]["url_original"]

    @url_original.setter
    def url_original(self, url):
        assert isinstance(url, str)
        self._response_internal["media_annotation"]["url_original"] = url

    @property
    def media_summary(self):
        return self._response_internal["media_annotation"]["media_summary"]

    @property
    def footprints(self):
        return self._response_internal["media_annotation"]["codes"]

    @property
    def width(self):
        return self._response_internal["media_annotation"]["w"]

    @width.setter
    def width(self, w):
        assert isinstance(w, int)
        self._response_internal["media_annotation"]["w"] = w

    @property
    def height(self):
        return self._response_internal["media_annotation"]["h"]

    @height.setter
    def height(self, h):
        assert isinstance(h, int)
        self._response_internal["media_annotation"]["h"] = h

    def get_timestamps_from_frames_ann(self):
        return sorted(self._tstamp2frameannsidx.keys())

    def get_timestamps(self, server=None):
        """Get timestamps from footprints. Option for querying on module name
        
        Args:
            server (str): module name
        
        Returns:
            List[float]: unique timestamps
        """
        tstamps = []
        for c in self._response_internal["media_annotation"]["codes"]:
            if server:
                if c["server"] != server:
                    continue
            tstamps.extend([round_float(t) for t in c["tstamps"]])
        return sorted(set(tstamps))

    # Modifiers

    def append_region(self, t, region):
        """Append a region given a timestamp. 

        If tstamp exists in frame_anns, appends region, otherwise, creates new ImageAnn

        Args:
            t (float): tstamp
            region (Region): region
        """
        assert isinstance(t, float)
        assert isinstance(region, type(Region()))
        log.debug(self._tstamp2frameannsidx)
        if t in self._tstamp2frameannsidx:
            log.debug(f"t: {t} in frame_anns, appending Region")
            frame_anns_idx = self._tstamp2frameannsidx[t]
            self._response_internal["media_annotation"]["frames_annotation"][frame_anns_idx]["regions"].append(region)
        else:
            log.debug(f"t: {t} NOT in frame_anns, appending ImageAnn")
            ia = ImageAnn(t=t, regions=[region])
            self._response_internal["media_annotation"]["frames_annotation"].append(ia)
            self._tstamp2frameannsidx[t] = len(self.frame_anns) - 1

    def append_regions(self, t, regions):
        """Append a list of regions given a timestamp

        If tstamp exists in frame_anns, appends regions, otherwise, creates new ImageAnns

        Args:
            t (float): tstamp
            region (Region): region
        """
        assert isinstance(t, float)
        assert isinstance(regions, list)
        for region in regions:
            assert isinstance(region, type(Region()))

        if t in self._tstamp2frameannsidx:
            log.debug(f"t: {t} in frame_anns, extending Regions")
            frame_anns_idx = self._tstamp2frameannsidx[t]
            self._response_internal["media_annotation"]["frames_annotation"][frame_anns_idx].extend(regions)
        else:
            log.debug(f"t: {t} NOT in frame_anns, appending ImageAnn")
            ia = ImageAnn(t=t, regions=regions)
            self._response_internal["media_annotation"]["frames_annotation"].append(ia)
            self._tstamp2frameannsidx[t] = len(self.frame_anns) - 1

    def append_footprint(self, footprint):
        """Append a footprint

        Args:
            footprint (Footprint): footprint
        """
        assert isinstance(footprint, type(Footprint()))
        self._response_internal["media_annotation"]["codes"].append(footprint)

    def append_track(self, video_ann):
        """Append a track to tracks_summary

        Args:
            video_ann (VideoAnn): track
        """
        assert isinstance(video_ann, type(VideoAnn()))
        self._response_internal["media_annotation"]["tracks_summary"].append(video_ann)

    def append_media_summary(self, video_ann):
        """Append a media_summary to media_summary section

        Args:
            video_ann (VideoAnn): media_summary
        """
        assert isinstance(video_ann, type(VideoAnn()))
        self._response_internal["media_annotation"]["media_summary"].append(video_ann)

    def sort_image_anns_by_timestamp(self):
        tmp = self._response_internal["media_annotation"]["frames_annotation"]
        self._response_internal["media_annotation"]["frames_annotation"] = sorted(
            tmp, key=lambda k: k["t"]
        )

    def sort_tracks_summary_by_timestamp(self):
        tmp = self._response_internal["media_annotation"]["tracks_summary"]
        self._response_internal["media_annotation"]["tracks_summary"] = sorted(
            tmp, key=lambda k: k["t1"]
        )

    def _init_from_request(self):
        """Construct from a request. 
        
        Request object has a field for "prev_response", if not None, 
        convert prev_response into a Response
        
        prev_responses can come in the following forms

            1) prev_response_url
                i) local file
                ii) remote file

            2) prev_response
                a) binary
                b) base64 binary string
                c) dict
        """
        log.debug("Constructing Response from Request")
        if self._request.prev_response_url:
            log.debug(f"Loading from prev_response_url: {self._request.prev_response_url}")
            try:
                file_retriever = FileRetriever(url=self._request.prev_response_url)
                f = file_retriever.download(return_filelike=True)
                self._request.prev_response = f.read().decode()
            except:
                log.error(traceback.format_exc())
        if self._request.prev_response:
            log.debug("Loading from prev_response")
            prev_response_dict = None
            if self._request.bin_encoding is True:
                log.debug("bin_encoding is True")
                io = AvroIO()
                if isinstance(self._request.prev_response, str):
                    log.debug("prev_response is base64 encoded binary")
                    prev_response_dict = io.decode(
                        self._request.prev_response, use_base64=True, binary_flag=True
                    )
                else:
                    log.debug("prev_response is in binary")
                    prev_response_dict = io.decode(
                        self._request.prev_response, use_base64=False, binary_flag=True
                    )
            else:
                if isinstance(self._request.prev_response, str):
                    log.debug("prev_response is a JSON str")
                    prev_response_dict = json.loads(self._request.prev_response)
                elif isinstance(self._request.prev_response, dict):
                    log.debug("prev_response is a dict")
                    prev_response_dict = self._request.prev_response
                else:
                    raise NotImplementedError("Intentionally not implemented--prev_response is not a JSON str or dict")

            if prev_response_dict is None:
                raise ValueError("error: prev_response_dict is None")
            try:
                self._response_internal = ResponseInternal(**prev_response_dict)
            except Exception as e:
                log.error("error unpacking prev_response_dict")
                log.error(traceback.format_exc())
        else:
            log.debug("No prev_response, constructing empty response_internal")
            self._response_internal = ResponseInternal()
        self.url = self._request.url

    def _init_tstamp2frameannsidx(self):
        """If there is a previous response, we want to store the tstamp 2 frame_anns idx dict
        """
        log.debug("Creating tstamp2frameannsidx")
        log.debug(f"prev_response frame_anns: {self.frame_anns}")
        for idx, image_ann in enumerate(self.frame_anns):
            self._tstamp2frameannsidx[round_float(image_ann["t"])] = idx
        log.debug(f"tstamp2frameannsidx: {self._tstamp2frameannsidx}")



import os
import json
import traceback

import glog as log
from flask import Flask, request, jsonify

from multivitamin.module import Module
from multivitamin.data import Request, Response


PORT = os.environ.get("PORT", 8888)


class WebServer(Flask):
    def __init__(self, modules, port=PORT):
        """Serves as the public interface for CV services through multivitamin

        It's role is to start the healthcheck endpoint and initiate the services

        Args:
            modules (list[Module]): list of concrete child implementation of CVModule
            input_comm (CommAPI): Concrete child implementation of CommAPI, called for pulling 
                                  responses to process
            output_comms (list[CommAPI]): List of concrete child implementations of CommAPI, 
                                          called for pushing responses to somewhere
        """
        if isinstance(modules, Module):
            modules = [modules]
        if not isinstance(modules, list):
            raise TypeError("modules is not a list of Module")
        if len(modules) == 0:
            raise TypeError("No modules was provided.")
        for m in modules:
            if not isinstance(m, Module):
                raise TypeError("Not all the modules are of type Module")
        self.modules_info = [{"name": x.name, "version": x.version} for x in modules]
        self.modules = modules
        self.port = port

        super().__init__(__name__)

        @self.route("/process", methods=["GET", "POST"])
        def process():
            """Entry point for starting an HTTP server
            """
            log.info("Pulling request")
            try:
                message = request.get_json(force=True)
                req = Request(message)
                response = Response(req)

                for module in self.modules:
                    log.info(f"Processing request for module: {module}")
                    response = module.process(response)
                    log.debug(
                        f"response.to_dict(): {json.dumps(response.to_dict(), indent=2)}"
                    )

                if req.bin_encoding:
                    return response.to_bytes()
                else:
                    return jsonify(response.to_dict())
            except Exception as e:
                log.error(e)
                log.error(traceback.print_exc())
                log.error(f"Error processing request {message}")

        @self.route("/health", methods=["GET"])
        def health_check():
            return jsonify(self.modules_info)

    def start(self):
        self.run(host="0.0.0.0", port=self.port)



import traceback
import glog as log
import requests as sender

from multivitamin.apis import CommAPI
from multivitamin.data import Response


class HTTPAPI(CommAPI):
    def pull(self, n=1):
        raise NotImplementedError()

    def push(self, responses, dst_url=None):
        """Pushes responses to a destination URL via the requests lib

        Args:
            responses (List[Response]): list of Response objects
            dst_url (str): url for POST endpoint
        """
        if not isinstance(responses, list):
            responses = [responses]

        log.debug(f"Pushing {len(responses)} items")
        for res in responses:
            assert isinstance(res, Response)
            if dst_url is None:
                dst_url = res.request.dst_url
            if dst_url:
                log.info(f"Pushing to {dst_url}")
                data = res.data
                log.info(f"data is of type {type(data)}")
                try:
                    ret = sender.post(dst_url, headers=self.auth_header, data=data)
                    log.info(f"requests.post(...) response: {ret}")
                except Exception as e:
                    log.error(e)
                    log.error(traceback.print_exc())
            else:
                log.info("No dst_url field in request. Not pushing response.")



import sys
import pims
import numpy as np
from .media_retriever import AbstractMediaRetriever, AbstractFramesIterator


class PIMSMediaRetriever(AbstractMediaRetriever):
    """A fileretriever for visual media."""

    def __init__(self, url=None):
        """Init MediaRetriever.

        Args:
            url (str | optional): The local or remote url to a file

        """
        super(AbstractMediaRetriever, self).__init__(url=url)

    @property
    def is_video(self):
        """Check if url is a video."""
        if self.content_type is None:
            return False
        return "video" in self.content_type

    @property
    def is_image(self):
        """Check if url is an image."""
        if self.content_type is None:
            return False
        return "image" in self.content_type and "gif" not in self.content_type

    def _create_video_capture(self):
        """Get the used video capture."""
        if self._cap is None:
            self._cap = pims.Video(self.url)

        return self._cap

    def _get_fps_from_video_capture(self):
        return self.video_capture.frame_rate

    def _get_num_frames(self):
        """Get the total number of frames."""
        return len(self.video_capture)

    def _get_video_frame_shape(self):
        """Get height by width by channels for frames."""
        shape = self.video_capture.frame_shape
        return shape

    def _get_frame_from_video(self, tstamp=0.0):
        """Return frame if image, or get frame with timestamp in seconds (w/ demicals).

        Note: iterating using get_frames_iterator is significantly faster

        """
        frame_idx = self.tstamp_to_frame_index(tstamp)
        ret = True
        frame = None
        try:
            frame = self.video_capture[frame_idx]
            frame = np.array(frame)[:, :, ::-1]  # RGB to BGR
        except ValueError:
            ret = False

        return ret, frame

    def _get_frames_iterator_class(self):
        return PIMSFramesIterator


class PIMSFramesIterator(AbstractFramesIterator):
    """Frames iterator object for videos."""

    def __init__(self,
                 video_cap,
                 video_fps,
                 sample_rate=100.0,
                 start_tstamp=0.0,
                 end_tstamp=sys.maxsize):
        """Frames iterator constructor.

        Args:
            video_cap (cv2.VideoCapture): video capture obj
            sample_rate (float): rate to sample video
            start_tstamp (float): start time for iterating
            end_tstamp (float): end time condition for iterating

        """
        super(PIMSFramesIterator, self).__init__(video_cap,
                                                 video_fps=video_fps,
                                                 sample_rate=sample_rate,
                                                 start_tstamp=start_tstamp,
                                                 end_tstamp=end_tstamp)

    def _move_cursor_to_tstamp(self):
        """Iterate over frames."""
        pass

    def _get_next_frame(self):
        """Get next frame."""
        ret = True
        frame = None
        tstamp = None
        try:
            frame_idx = round(self.cur_tstamp * self.cap.frame_rate)
            frame = self.cap[frame_idx]
            frame = np.array(frame)[:, :, ::-1]
            self.cur_tstamp += self.period
            tstamp = frame_idx / self.cap.frame_rate
            return frame, self._round_tstamp(tstamp)
        except ValueError:
            ret = False

        return ret, frame, tstamp



from multivitamin.media import MediaRetriever

from tabulate import tabulate
from datetime import datetime
import random
from tqdm import tqdm

print("SPEED TEST!!!")

VIDEO_URL = "https://s3.amazonaws.com/video-ann-testing/NHL_GAME_VIDEO_NJDMTL_M2_NATIONAL_20180401_1520698069177.t.mp4"
VIDEO_CODEC_PROB_1 = "https://s3.amazonaws.com/video-ann/538_Pelicans+vs+Thunder+11%3A5-fhj713lbrhi.30-31.mp4"
VIDEO_CODEC_PROB_2 = "https://s3.amazonaws.com/gumgum-sports-analyst-data/media-files/Replay%20Video%20Capture_2018-11-16_11.52.51-2816an1tb0v.mp4"
VIDEO_CODEC_PROB_3 = "https://s3.amazonaws.com/gumgum-sports-analyst-data/media-files/1%3A3%20Houston%20Rockets%20at%20Golden%20State%20Warriors-6tgm4my1dr6.mp4"

VIDEO_URLS = [VIDEO_URL, VIDEO_CODEC_PROB_1, VIDEO_CODEC_PROB_2, VIDEO_CODEC_PROB_3]


def create_media_retrievers(url):
    efficient_mr = MediaRetriever(VIDEO_URL)
    fast_mr = MediaRetriever(VIDEO_URL, limitation="cpu")
    return efficient_mr, fast_mr


def _benchmark_get_frame(mrs, num_samples=100, num_tests=100):
    averaged_results = []
    random_samples = np.random.rand(num_samples)
    for mr in mrs:
        random_tstamps = random_samples * mr.length
        results = []
        for _ in tqdm(range(num_tests), desc="get frame"):
            start = datetime.now()
            for tstamp in random_samples:
                mr.get_frame(tstamp)
            end = datetime.now()
            x = end - start
            results.append(x.total_seconds())
        averaged_results.append(np.average(results))
    return averaged_results


def _benchmark_frames_iterator(mrs, sample_rate, num_samples=100, num_tests=100):
    averaged_results = []
    random_sample = random.random()
    for mr in mrs:
        random_start = random_sample * mr.length
        results = []
        for _ in tqdm(range(num_tests), desc="iterator test"):
            start = datetime.now()
            for idx, _ in enumerate(
                mr.get_frames_iterator(
                    sample_rate=sample_rate, start_tstamp=random_start
                )
            ):
                if idx >= num_samples:
                    break
            end = datetime.now()
            x = end - start
            results.append(x.total_seconds())
        averaged_results.append(np.average(results))
    return averaged_results


rs = []
for url in VIDEO_URLS:
    mrs = create_media_retrievers(url)
    headers = ["Test Type", "Efficient", "Fast"]

    test_order = [
        "get_frame",
        "iterator (sample_rate=100)",
        "iterator (sample_rate=2)",
        "iterator (sample_rate=0.2)",
    ]
    get_frame_results = _benchmark_get_frame(mrs, num_tests=1)
    iterator_results1 = _benchmark_frames_iterator(mrs, 100, num_tests=5)
    iterator_results2 = _benchmark_frames_iterator(mrs, 2, num_tests=5)
    iterator_results3 = _benchmark_frames_iterator(mrs, 0.2, num_tests=5)

    results = list(
        zip(
            test_order,
            get_frame_results,
            iterator_results1,
            iterator_results2,
            iterator_results3,
        )
    )
    rs.append(results)

print("\n" * 4)
for url, results in zip(VIDEO_URLS, rs):
    print("\nRESULTS ON: " + url)
    print(tabulate(results, headers=headers))



from abc import ABC, abstractmethod
import json

import glog as log

from multivitamin.data import Response
from multivitamin.data.response.dtypes import Footprint
from multivitamin.data.response.utils import get_current_time
from multivitamin.module.codes import Codes
from multivitamin.module.utils import convert_props_to_pandas_query


class Module(ABC):
    def __init__(
        self,
        server_name,
        version,
        prop_type=None,
        prop_id_map=None,
        module_id_map=None
    ):
        """Abstract base class that defines interface inheritance

        ImageModule, PropertiesModule

        Handles processing of request and previous response
        """
        self.name = server_name
        self.version = version
        self.prop_type = prop_type
        self.prop_id_map = prop_id_map
        self.module_id_map = module_id_map
        self.prev_pois = None
        self.code = Codes.SUCCESS
        self.prev_regions_of_interest_count = 0
        self.tstamps_processed = []

    def set_prev_props_of_interest(self, pois):
        """If this Module is meant to be one in a sequence of Modules and is
        looking for a particular set of properties of interest from the
        previous module.

        E.g.
            [
                {"property_type":"object", "value":"face"},
                {"value":"car"}
            ]

            is equivalent to

            '(property_type == "object") & (value == "face") | (value == "car")'

        Args:
            pois (list[dict]): previous properties of interest
        """
        assert isinstance(pois, list)
        for poi in pois:
            assert isinstance(poi, dict)

        self.prev_pois = pois
        self.prev_pois_bool_exp = convert_props_to_pandas_query(pois)
        log.info(
            "Setting previous properties of interest: "
            f"{json.dumps(pois, indent=2)}"
        )
        log.info(f"POIs for {self}: {self.prev_pois_bool_exp}")

    def get_prev_props_of_interest(self):
        """Getter for properties of interest

        Returns:
            list[dict]: prev props of interest
        """
        return self.prev_pois

    @abstractmethod
    def process(self, response):
        """Abstract method, public entry point for procesing a response

        Args:
            response (Response): response
        """
        assert isinstance(response, Response)
        self.tstamps_processed = []
        self.code = Codes.SUCCESS
        self.response = response
        self.request = response.request

    def update_and_return_response(self):
        """Update footprints, moduleID, propertyIDs

        Note: to be moved into aigumgum

        Returns:
            Response: output response
        """
        log.info(
            f"{self} updating and returning response with code:"
            f" {self.code.name}"
        )
        num_footprints = len(self.response.footprints)
        time = get_current_time()
        log.debug("Appending footprints")
        self.response.append_footprint(
            Footprint(
                code=self.code.name,
                server=self.name,
                date=time,
                ver=self.version,
                id="{}{}".format(time, num_footprints + 1),
                tstamps=self.tstamps_processed,
                num_images_processed=len(self.tstamps_processed)
            )
        )
        self._update_ids()
        return self.response

    def _update_ids(self):
        """Internal method for updating property id and module id
           in frame_anns and tracks
        """
        for image_ann in self.response.frame_anns:
            for region in image_ann["regions"]:
                for prop in region["props"]:
                    self._update_property_id(prop)
                    self._update_module_id(prop)

        for video_ann in self.response.tracks:
            for prop in video_ann["props"]:
                self._update_property_id(prop)
                self._update_module_id(prop)

        for video_ann in self.response.media_summary:
            for prop in video_ann["props"]:
                self._update_property_id(prop)
                self._update_module_id(prop)

    def _update_property_id(self, prop):
        """Internal method for updating property id in a property
        """
        log.debug("Updating property ids")
        if self.prop_id_map is None:
            return
        if prop["property_id"] == 0:
            value = prop.get("value", "")
            prop["property_id"] = self.prop_id_map.get(value, 0)

    def _update_module_id(self, prop):
        """Internal method for updating module id in a property
        """
        log.debug("Updating module ids")
        if self.module_id_map is None:
            return
        if prop["module_id"] == 0:
            server = prop.get("server", "")
            prop["module_id"] = self.module_id_map.get(server, 0)

    def __repr__(self):
        return f"{self.__class__}(name={self.name}, version={self.version})"
